from unittest import TestCase

from lxml import etree

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)
from pcs_test.tools.xml import etree_to_str

from pcs.common import report_codes
from pcs.lib.cib.resource import hierarchy
from pcs.lib.cib.tools import IdProvider


def _resource(cib, id_):
    return cib.find(f".//*[@id='{id_}']")

def _resources(cib, *ids):
    return [_resource(cib, id_) for id_ in ids]

class ValidateMoveResourcesToGroupMixin():
    def setUp(self):
        # pylint does not know this mixin goes to TestCase
        # pylint: disable=invalid-name
        self.cib = etree.fromstring("""
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                </group>
                <group id="GX">
                    <primitive id="RGX" />
                </group>
                <primitive id="R1" />
                <primitive id="R2" />
                <primitive id="R3" />
                <clone id="RC1-clone">
                    <primitive id="RC1" />
                </clone>
                <bundle id="RB1-bundle">
                    <primitive id="RB1">
                        <meta_attributes id="RB1-meta_attributes" />
                    </primitive>
                </bundle>
            </resources>
        """)

    def test_no_resources_specified(self):
        assert_report_item_list_equal(
            self._validate("G", []),
            [
                fixture.error(report_codes.CANNOT_GROUP_RESOURCE_NO_RESOURCES),
            ]
        )

    def test_group_is_not_group(self):
        assert_report_item_list_equal(
            self._validate("RB1-meta_attributes", ["R1"]),
            [
                fixture.error(
                    report_codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="RB1-meta_attributes",
                    expected_types=["group"],
                    current_type="meta_attributes",
                ),
            ]
        )

    def test_resources_are_not_primitives(self):
        assert_report_item_list_equal(
            self._validate("G", ["RC1-clone", "R1", "RB1-bundle"]),
            [
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RC1-clone",
                    resource_type="clone",
                ),
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RB1-bundle",
                    resource_type="bundle",
                ),
            ]
        )

    def test_resources_are_in_clones_etc(self):
        assert_report_item_list_equal(
            self._validate("G", ["RC1", "R1", "RB1"]),
            [
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RC1",
                    resource_type="clone",
                ),
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RB1",
                    resource_type="bundle",
                ),
            ]
        )

    def test_resources_already_in_the_group(self):
        assert_report_item_list_equal(
            self._validate("G", ["RG2", "R1", "RG1"]),
            [
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_ALREADY_IN_THE_GROUP,
                    resource_list=["RG1", "RG2"],
                    group_id="G",
                ),
            ]
        )

    def test_allow_moving_resources_in_a_group_if_adjacent(self):
        assert_report_item_list_equal(
            self._validate("G", ["RG2", "R1"], "RG1"),
            [
            ]
        )

    def test_adjacent_resource_not_in_the_group(self):
        assert_report_item_list_equal(
            self._validate("G", ["R1"], "R2"),
            [
                fixture.error(
                    report_codes
                        .CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_NOT_IN_GROUP
                    ,
                    adjacent_resource_id="R2",
                    group_id="G",
                ),
            ]
        )

    def test_adjacent_resource_in_another_group(self):
        assert_report_item_list_equal(
            self._validate("G", ["R1"], "RGX"),
            [
                fixture.error(
                    report_codes
                        .CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_NOT_IN_GROUP
                    ,
                    adjacent_resource_id="RGX",
                    group_id="G",
                ),
            ]
        )

    def test_adjacent_resource_to_be_grouped(self):
        assert_report_item_list_equal(
            self._validate("G", ["RG1"], "RG1"),
            [
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_NEXT_TO_ITSELF,
                    resource_id="RG1",
                ),
            ]
        )

    def test_resources_specified_twice(self):
        assert_report_item_list_equal(
            self._validate("G", ["R3", "R2", "R1", "R2", "R1"]),
            [
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_MORE_THAN_ONCE,
                    resource_list=["R1", "R2"],
                ),
            ]
        )


class ValidateMoveResourcesToGroupByElements(
    ValidateMoveResourcesToGroupMixin, TestCase
):
    def _resource(self, id_):
        return _resource(self.cib, id_)

    def _resources(self, ids):
        return _resources(self.cib, *ids)

    def _validate(self, group, resources, adjacent=None):
        return hierarchy.ValidateMoveResourcesToGroupByElements(
            self._resource(group),
            self._resources(resources),
            self._resource(adjacent) if adjacent else None,
        ).validate()

    def test_resources_are_not_resources(self):
        # The validator expects to get resource elements. So this report is
        # not the best, but at least the validator detects the problem.
        # Validation using IDs provides better reporting in this case.
        assert_report_item_list_equal(
            self._validate("G", ["RB1-meta_attributes"]),
            [
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RB1-meta_attributes",
                    resource_type="meta_attributes",
                ),
            ]
        )


class ValidateMoveResourcesToGroupByIds(
    ValidateMoveResourcesToGroupMixin, TestCase
):
    def _resource(self, id_):
        # this is for inheritance
        # pylint: disable=no-self-use
        return id_

    def _resources(self, ids):
        # this is for inheritance
        # pylint: disable=no-self-use
        return ids

    def _validate(self, group, resources, adjacent=None):
        return hierarchy.ValidateMoveResourcesToGroupByIds(
            self._resource(group),
            self._resources(resources),
            self._resource(adjacent) if adjacent else None,
        ).validate(
            self.cib,
            IdProvider(self.cib),
        )

    def test_new_group_not_valid_id(self):
        assert_report_item_list_equal(
            self._validate("1Gr:oup", ["R1"]),
            [
                fixture.error(
                    report_codes.INVALID_ID,
                    id="1Gr:oup",
                    id_description="group name",
                    is_first_char=True,
                    invalid_character="1",
                ),
                fixture.error(
                    report_codes.INVALID_ID,
                    id="1Gr:oup",
                    id_description="group name",
                    is_first_char=False,
                    invalid_character=":",
                ),
            ]
        )

    def test_missing_resources_specified(self):
        assert_report_item_list_equal(
            self._validate("G", ["RX1", "RX2"]),
            [
                fixture.report_not_found("RX1", context_type="resources"),
                fixture.report_not_found("RX2", context_type="resources"),
            ]
        )

    def test_resources_are_not_resources(self):
        assert_report_item_list_equal(
            self._validate("G", ["RB1-meta_attributes"]),
            [
                fixture.error(
                    report_codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="RB1-meta_attributes",
                    expected_types=[
                        "clone", "master", "group", "primitive", "bundle"
                    ],
                    current_type="meta_attributes",
                ),
            ]
        )

    def test_adjacent_resource_new_group(self):
        assert_report_item_list_equal(
            self._validate("G-new", ["R1"], "R2"),
            [
                fixture.error(
                    report_codes
                        .CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_FOR_NEW_GROUP
                    ,
                    adjacent_resource_id="R2",
                    group_id="G-new",
                ),
            ]
        )

    def test_adjacent_resource_doesnt_exist(self):
        assert_report_item_list_equal(
            self._validate("G", ["R1"], "RX"),
            [
                fixture.error(
                    report_codes
                        .CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_NOT_IN_GROUP
                    ,
                    adjacent_resource_id="RX",
                    group_id="G",
                ),
            ]
        )




class MoveResourcesToGroup(TestCase):
    cib_resources = """
        <resources>
            <group id="G">
                <primitive id="RG1" />
                <primitive id="RG2" />
                <primitive id="RG3" />
            </group>
            <primitive id="R1" />
            <primitive id="R2" />
            <primitive id="R3" />
        </resources>
    """

    def test_no_adjacent(self):
        cib_before = self.cib_resources
        cib_after = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                    <primitive id="RG3" />
                    <primitive id="R3" />
                    <primitive id="R1" />
                </group>
                <primitive id="R2" />
            </resources>
        """
        cib = etree.fromstring(cib_before)
        hierarchy.move_resources_to_group(
            _resource(cib, "G"),
            _resources(cib, "R3", "R1")
        )
        assert_xml_equal(cib_after, etree_to_str(cib))

    def test_after_adjacent(self):
        cib_before = self.cib_resources
        cib_after = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="R3" />
                    <primitive id="R1" />
                    <primitive id="RG3" />
                    <primitive id="RG2" />
                </group>
                <primitive id="R2" />
            </resources>
        """
        cib = etree.fromstring(cib_before)
        hierarchy.move_resources_to_group(
            _resource(cib, "G"),
            _resources(cib, "R3", "R1", "RG3"),
            adjacent_resource=_resource(cib, "RG1")
        )
        assert_xml_equal(cib_after, etree_to_str(cib))

    def test_after_adjacent_last(self):
        cib_before = self.cib_resources
        cib_after = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG3" />
                    <primitive id="R3" />
                    <primitive id="R1" />
                    <primitive id="RG2" />
                </group>
                <primitive id="R2" />
            </resources>
        """
        cib = etree.fromstring(cib_before)
        hierarchy.move_resources_to_group(
            _resource(cib, "G"),
            _resources(cib, "R3", "R1", "RG2"),
            adjacent_resource=_resource(cib, "RG3")
        )
        assert_xml_equal(cib_after, etree_to_str(cib))

    def test_before_adjacent(self):
        cib_before = self.cib_resources
        cib_after = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="R3" />
                    <primitive id="R1" />
                    <primitive id="RG3" />
                    <primitive id="RG2" />
                </group>
                <primitive id="R2" />
            </resources>
        """
        cib = etree.fromstring(cib_before)
        hierarchy.move_resources_to_group(
            _resource(cib, "G"),
            _resources(cib, "R3", "R1", "RG3"),
            adjacent_resource=_resource(cib, "RG2"),
            put_after_adjacent=False
        )
        assert_xml_equal(cib_after, etree_to_str(cib))

    def test_before_adjacent_first(self):
        cib_before = self.cib_resources
        cib_after = """
            <resources>
                <group id="G">
                    <primitive id="R3" />
                    <primitive id="R1" />
                    <primitive id="RG3" />
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                </group>
                <primitive id="R2" />
            </resources>
        """
        cib = etree.fromstring(cib_before)
        hierarchy.move_resources_to_group(
            _resource(cib, "G"),
            _resources(cib, "R3", "R1", "RG3"),
            adjacent_resource=_resource(cib, "RG1"),
            put_after_adjacent=False
        )
        assert_xml_equal(cib_after, etree_to_str(cib))

    def test_move_from_another_group(self):
        # pylint: disable=no-self-use
        cib_before = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                    <primitive id="RG3" />
                </group>
                <group id="X">
                    <primitive id="R1" />
                    <primitive id="R2" />
                </group>
                <primitive id="R3" />
            </resources>
        """
        cib_after = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                    <primitive id="RG3" />
                    <primitive id="R2" />
                </group>
                <group id="X">
                    <primitive id="R1" />
                </group>
                <primitive id="R3" />
            </resources>
        """
        cib = etree.fromstring(cib_before)
        hierarchy.move_resources_to_group(
            _resource(cib, "G"),
            _resources(cib, "R2"),
        )
        assert_xml_equal(cib_after, etree_to_str(cib))

    def test_move_last_from_group(self):
        # pylint: disable=no-self-use
        cib_before = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                    <primitive id="RG3" />
                </group>
                <group id="X">
                    <primitive id="R1" />
                </group>
                <primitive id="R3" />
            </resources>
        """
        cib_after = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                    <primitive id="RG3" />
                    <primitive id="R1" />
                </group>
                <primitive id="R3" />
            </resources>
        """
        cib = etree.fromstring(cib_before)
        hierarchy.move_resources_to_group(
            _resource(cib, "G"),
            _resources(cib, "R1"),
        )
        assert_xml_equal(cib_after, etree_to_str(cib))

    def _assert_move_last_from_cloned_group(self, clone_type):
        # pylint: disable=no-self-use
        cib_before = f"""
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                    <primitive id="RG3" />
                </group>
                <{clone_type} id="X-{clone_type}">
                    <group id="X">
                        <primitive id="R1" />
                    </group>
                </{clone_type}>
                <primitive id="R3" />
            </resources>
        """
        cib_after = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                    <primitive id="RG3" />
                    <primitive id="R1" />
                </group>
                <primitive id="R3" />
            </resources>
        """
        cib = etree.fromstring(cib_before)
        hierarchy.move_resources_to_group(
            _resource(cib, "G"),
            _resources(cib, "R1"),
        )
        assert_xml_equal(cib_after, etree_to_str(cib))

    def test_move_last_from_cloned_group(self):
        self._assert_move_last_from_cloned_group("clone")

    def test_move_last_from_mastered_group(self):
        self._assert_move_last_from_cloned_group("master")



from pcs import (
    constraint,
    usage,
)
from pcs.cli.common.routing import create_router
import pcs.cli.constraint_colocation.command as colocation_command
from pcs.cli.constraint_ticket import command as ticket_command


constraint_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: usage.constraint(argv),
        "location": constraint.constraint_location_cmd,
        "order": constraint.constraint_order_cmd,
        "ticket": create_router(
            {
                "set": ticket_command.create_with_set,
                "add": ticket_command.add,
                "delete": ticket_command.remove,
                "remove": ticket_command.remove,
                "show": ticket_command.show,
            },
            ["constraint", "ticket"],
            default_cmd="show"
        ),
        "colocation": create_router(
            {
                "add": constraint.colocation_add,
                "remove": constraint.colocation_rm,
                "delete": constraint.colocation_rm,
                "set": colocation_command.create_with_set,
                "show": colocation_command.show,
            },
            ["constraint", "colocation"],
            default_cmd="show"
        ),
        "remove": constraint.constraint_rm,
        "delete": constraint.constraint_rm,
        "show": constraint.constraint_show,
        "list": constraint.constraint_show,
        "ref": constraint.constraint_ref,
        "rule": constraint.constraint_rule,
    },
    ["constraint"],
    default_cmd="list"
)



from pcs_test.tools.command_env.mock_runner import Call as RunnerCall

from pcs import settings

class SbdShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def list_watchdogs(
        self,
        output,
        name="runner.sbd.list_watchdogs",
        stderr="",
        returncode=0,
        instead=None,
        before=None
    ):
        self.__calls.place(
            name,
            RunnerCall(
                "{} query-watchdog".format(settings.sbd_binary),
                stdout=output,
                stderr=stderr,
                returncode=returncode,
            ),
            before=before,
            instead=instead
        )



def get_resources_failcounts(cib_status):
    # pylint: disable=too-many-locals
    """
    List all resources failcounts
    Return a dict {
        "node": string -- node name,
        "resource": string -- resource id,
        "clone_id": string -- resource clone id or None,
        "operation": string -- operation name,
        "interval": string -- operation interval,
        "fail_count": "INFINITY" or int -- fail count,
        "last_failure": int -- last failure timestamp,
    }

    etree cib_status -- status element of the CIB
    """
    failcounts = []
    for node_state in cib_status.findall("node_state"):
        node_name = node_state.get("uname")

        # Pair fail-counts with last-failures.
        # failures_info = {
        #         failure_name: {"fail_count": count, "last-failure": timestamp}
        #     }
        failures_info = {}
        for nvpair in node_state.findall(
            "transient_attributes/instance_attributes/nvpair"
        ):
            name = nvpair.get("name")
            for part in ("fail-count-", "last-failure-"):
                if name.startswith(part):
                    failure_name = name[len(part):]
                    if failure_name not in failures_info:
                        failures_info[failure_name] = {}
                    failures_info[failure_name][part[:-1]] = nvpair.get("value")
                    break

        if not failures_info:
            continue
        for failure_name, failure_data in failures_info.items():
            resource, clone_id, operation, interval = _parse_failure_name(
                failure_name
            )
            fail_count = failure_data.get("fail-count", "0").upper()
            if fail_count != "INFINITY":
                try:
                    fail_count = int(fail_count)
                except ValueError:
                    # There are failures we just do not know how many. If we set
                    # fail_count = 0, no failures would be recorded.
                    fail_count = 1
            try:
                last_failure = int(failure_data.get("last-failure", "0"))
            except ValueError:
                last_failure = 0
            failcounts.append({
                "node": node_name,
                "resource": resource,
                "clone_id": clone_id,
                "operation": operation,
                "interval": interval,
                "fail_count": fail_count,
                "last_failure": last_failure,
            })
    return failcounts

def _parse_failure_name(name):
    # failure_name looks like this:
    # <resource_name>[:<clone_id>]#<operation>_<interval>
    # resource name is an id so it cannot contain # nor :
    resource_clone, operation_interval = name.split("#", 1)
    if ":" in resource_clone:
        resource, clone = resource_clone.split(":", 1)
    else:
        resource, clone = resource_clone, None
    operation, interval = operation_interval.rsplit("_", 1)
    return resource, clone, operation, interval

def filter_resources_failcounts(
    failcounts, resource=None, node=None, operation=None, interval=None
):
    return [
        failure for failure in failcounts
        if (
            (node is None or failure["node"] == node)
            and
            (resource is None or failure["resource"] == resource)
            and
            (operation is None or failure["operation"] == operation)
            and
            # 5 != "5", failure["interval"] is a string already
            (interval is None or failure["interval"] == str(interval))
        )
    ]



# pylint: disable=no-member
from functools import partial
from unittest import mock, TestCase

from pcs_test.tier0.lib.commands.remote_node.fixtures_add import(
    EnvConfigMixin,
    REPORTS as FIXTURE_REPORTS,
    EXTRA_REPORTS as FIXTURE_EXTRA_REPORTS,
    FAIL_HTTP_KWARGS,
    fixture_reports_not_live_cib,
    fixture_reports_new_node_unreachable,
)
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc

from pcs import settings
from pcs.common import file_type_codes, report_codes
from pcs.common.file import RawFileError
from pcs.common.host import Destination
from pcs.lib.commands.remote_node import node_add_remote as node_add_remote_orig


NODE_NAME = "node-name"
NODE_ADDR_PCSD = "node-addr"
NODE_DEST_LIST = [Destination(NODE_ADDR_PCSD, 2224)]
NODE_1 = "node-1"
NODE_1_DEST_LIST = [Destination("node-1-addr", 2224)]
NODE_2 = "node-2"
NODE_2_DEST_LIST = [Destination("node-2-addr", 2224)]
KNOWN_HOSTS_DESTS = {
    NODE_NAME: NODE_DEST_LIST,
    NODE_1: NODE_1_DEST_LIST,
    NODE_2: NODE_2_DEST_LIST,
}

def node_add_remote(
    env, node_name=None, node_addr=None, operations=None, meta_attributes=None,
    instance_attributes=None, no_node_addr=False, **kwargs
):
    operations = operations or []
    meta_attributes = meta_attributes or {}
    instance_attributes = instance_attributes or {}
    node_name = node_name or NODE_NAME
    node_addr = node_addr or (None if no_node_addr else "remote-host")

    node_add_remote_orig(
        env, node_name, node_addr, operations, meta_attributes,
        instance_attributes, **kwargs
    )

class LocalConfig(EnvConfigMixin):
    def load_cluster_configs(self, cluster_node_list):
        (self.config
            .runner.cib.load()
            .corosync_conf.load(node_name_list=cluster_node_list)
            .runner.pcmk.load_agent(agent_name="ocf:pacemaker:remote")
        )

get_env_tools = partial(get_env_tools, local_extensions={"local": LocalConfig})

REPORTS = (FIXTURE_REPORTS
    .adapt("authkey_distribution_started", node_list=[NODE_NAME])
    .adapt("authkey_distribution_success", node=NODE_NAME)
    .adapt("pcmk_remote_start_enable_started", node_list=[NODE_NAME])
    .adapt("pcmk_remote_enable_success", node=NODE_NAME)
    .adapt("pcmk_remote_start_success", node=NODE_NAME)
)
EXTRA_REPORTS = (FIXTURE_EXTRA_REPORTS.adapt_multi(
    [
        "manage_services_connection_failed",
        "manage_services_connection_failed_warn",
        "check_availability_connection_failed",
        "check_availability_connection_failed_warn",
        "put_file_connection_failed",
        "put_file_connection_failed_warn",
        "pcmk_remote_enable_failed",
        "pcmk_remote_enable_failed_warn",
        "pcmk_remote_start_failed",
        "pcmk_remote_start_failed_warn",
        "authkey_distribution_failed",
        "authkey_distribution_failed_warn",
    ],
    node=NODE_NAME
))


FIXTURE_RESOURCES_TEMPLATE = """
    <resources>
        <primitive class="ocf" id="node-name" provider="pacemaker"
            type="remote"
        >
            <instance_attributes id="node-name-instance_attributes">
                <nvpair
                    id="node-name-instance_attributes-server"
                    name="server" value="{server}"
                />
            </instance_attributes>
            <operations>
                <op id="node-name-migrate_from-interval-0s"
                    interval="0s" name="migrate_from" timeout="60"
                />
                <op id="node-name-migrate_to-interval-0s"
                    interval="0s" name="migrate_to" timeout="60"
                />
                <op id="node-name-monitor-interval-60s"
                    interval="60s" name="monitor" timeout="30"
                />
                <op id="node-name-reload-interval-0s"
                  interval="0s" name="reload" timeout="60"
                />
                <op id="node-name-start-interval-0s"
                    interval="0s" name="start" timeout="60"
                />
                <op id="node-name-stop-interval-0s"
                    interval="0s" name="stop" timeout="60"
                />
            </operations>
        </primitive>
    </resources>
"""
FIXTURE_RESOURCES = FIXTURE_RESOURCES_TEMPLATE.format(server="remote-host")

class AddRemote(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.env.set_known_hosts_dests(KNOWN_HOSTS_DESTS)

    def _config_success_base(self):
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
                ],
            )
            .local.get_host_info(NODE_NAME, NODE_DEST_LIST)
            .local.push_existing_authkey_to_remote(NODE_NAME, NODE_DEST_LIST)
            .local.run_pacemaker_remote(NODE_NAME, NODE_DEST_LIST)
            .env.push_cib(resources=FIXTURE_RESOURCES)
        )

    def test_success_base(self):
        self._config_success_base()
        node_add_remote(self.env_assist.get_env())
        self.env_assist.assert_reports(REPORTS)

    def test_success_base_addr_same_as_name(self):
        #validation and creation of resource is covered in resource create tests
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
                ],
            )
            .local.get_host_info(NODE_NAME, NODE_DEST_LIST)
            .local.push_existing_authkey_to_remote(NODE_NAME, NODE_DEST_LIST)
            .local.run_pacemaker_remote(NODE_NAME, NODE_DEST_LIST)
            .env.push_cib(
                resources=FIXTURE_RESOURCES_TEMPLATE.format(
                    server=NODE_NAME
                )
            )
        )
        node_add_remote(self.env_assist.get_env(), node_addr=NODE_NAME)
        self.env_assist.assert_reports(REPORTS)

    def test_node_name_conflict_report_is_unique(self):
        (self.config
            .runner.cib.load(
                resources="""
                    <resources>
                        <primitive class="ocf" id="node-name"
                            provider="pacemaker" type="Dummy"
                        />
                    </resources>
                """
            )
            .corosync_conf.load(node_name_list=[NODE_1, NODE_2])
            .runner.pcmk.load_agent(agent_name="ocf:pacemaker:remote")
        )

        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(self.env_assist.get_env()),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.ID_ALREADY_EXISTS,
                    id=NODE_NAME,
                )
            ]
        )

    @mock.patch("pcs.lib.commands.remote_node.generate_binary_key")
    def test_success_generated_authkey(self, generate_binary_key):
        generate_binary_key.return_value = b"password"
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
                ],
            )
            .local.get_host_info(NODE_NAME, NODE_DEST_LIST)
            .local.authkey_exists(return_value=False)
            .local.distribute_authkey(
                communication_list=[
                    dict(label=NODE_1, dest_list=NODE_1_DEST_LIST),
                    dict(label=NODE_2, dest_list=NODE_2_DEST_LIST),
                    dict(label=NODE_NAME, dest_list=NODE_DEST_LIST),
                ],
                pcmk_authkey_content=generate_binary_key.return_value,
            )
            .local.run_pacemaker_remote(NODE_NAME, NODE_DEST_LIST)
            .env.push_cib(resources=FIXTURE_RESOURCES)
        )
        node_add_remote(self.env_assist.get_env())
        generate_binary_key.assert_called_once_with(random_bytes_count=256)
        self.env_assist.assert_reports(
            REPORTS
                .adapt(
                    "authkey_distribution_started",
                    node_list=[NODE_1, NODE_2, NODE_NAME]
                )
                .copy(
                    "authkey_distribution_success",
                    "authkey_distribution_success_node1",
                    node=NODE_1,
                )
                .copy(
                    "authkey_distribution_success",
                    "authkey_distribution_success_node2",
                    node=NODE_2,
                )
        )

    def test_new_offline(self):
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(
                        label=NODE_NAME,
                        dest_list=NODE_DEST_LIST,
                        **FAIL_HTTP_KWARGS
                    )
                ],
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(self.env_assist.get_env()),
            []
        )
        self.env_assist.assert_reports([
            fixture.error(
                report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                force_code=report_codes.SKIP_OFFLINE_NODES,
                node=NODE_NAME,
                command="remote/check_auth",
                reason="Could not resolve host",
            )
        ])

    def test_can_skip_new_offline(self):
        pcmk_authkey_content = b"password"
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(
                        label=NODE_NAME,
                        dest_list=NODE_DEST_LIST,
                        **FAIL_HTTP_KWARGS
                    )
                ],
            )
            .local.authkey_exists(return_value=True)
            .local.open_authkey(pcmk_authkey_content)
            .env.push_cib(resources=FIXTURE_RESOURCES)
        )
        node_add_remote(self.env_assist.get_env(), skip_offline_nodes=True)
        self.env_assist.assert_reports(
            fixture_reports_new_node_unreachable(NODE_NAME, omitting=True)
        )

    @mock.patch("pcs.lib.commands.remote_node.generate_binary_key")
    def test_can_skip_all_offline(self, generate_binary_key):
        generate_binary_key.return_value = b"password"
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(
                        label=NODE_NAME,
                        dest_list=NODE_DEST_LIST,
                        **FAIL_HTTP_KWARGS
                    )
                ],
            )
            .local.authkey_exists(return_value=False)
            .local.distribute_authkey(
                communication_list=[
                    dict(label=NODE_1, dest_list=NODE_1_DEST_LIST),
                    dict(label=NODE_2, dest_list=NODE_2_DEST_LIST),
                ],
                pcmk_authkey_content=generate_binary_key.return_value,
                **FAIL_HTTP_KWARGS
            )
            .env.push_cib(resources=FIXTURE_RESOURCES)
        )
        node_add_remote(self.env_assist.get_env(), skip_offline_nodes=True)
        self.env_assist.assert_reports(
            fixture_reports_new_node_unreachable(NODE_NAME, omitting=True)
            + [
            fixture.info(
                report_codes.FILES_DISTRIBUTION_STARTED,
                file_list=["pacemaker authkey"],
                node_list=[NODE_1, NODE_2],
            ),
            fixture.warn(
                report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                node=NODE_1,
                command="remote/put_file",
                reason="Could not resolve host",
            ),
            fixture.warn(
                report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                node=NODE_2,
                command="remote/put_file",
                reason="Could not resolve host",
            ),
        ])

    def test_fails_when_remote_node_is_not_prepared(self):
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
                ],
            )
            .local.get_host_info(
                NODE_NAME,
                NODE_DEST_LIST,
                output=dict(
                    services=dict(
                        pacemaker_remote=dict(
                            installed=False, enabled=False, running=False
                        ),
                        pacemaker=dict(
                            installed=True, enabled=False, running=True
                        ),
                        corosync=dict(
                            installed=True, enabled=False, running=True
                        ),
                    ),
                    cluster_configuration_exists=True,
                )
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(self.env_assist.get_env()),
            []
        )
        self.env_assist.assert_reports([
            fixture.error(
                report_codes.SERVICE_NOT_INSTALLED,
                node=NODE_NAME,
                service_list=["pacemaker_remote"],
            ),
            fixture.error(
                report_codes.HOST_ALREADY_IN_CLUSTER_SERVICES,
                host_name=NODE_NAME,
                service_list=["corosync", "pacemaker"],
            ),
            fixture.error(
                report_codes.HOST_ALREADY_IN_CLUSTER_CONFIG,
                host_name=NODE_NAME,
            ),
        ])

    def test_fails_when_remote_node_returns_invalid_output(self):
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
                ],
            )
            .local.get_host_info(
                NODE_NAME, NODE_DEST_LIST, output="INVALID_OUTPUT"
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(self.env_assist.get_env()),
            []
        )
        self.env_assist.assert_reports([
            fixture.error(
                report_codes.INVALID_RESPONSE_FORMAT,
                node=NODE_NAME,
            )
        ])

    def test_open_failed(self):
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
                ],
            )
            .local.get_host_info(NODE_NAME, NODE_DEST_LIST)
            .local.authkey_exists(return_value=True)
            .local.open_authkey(fail=True)
        )

        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(
                self.env_assist.get_env(),
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.FILE_IO_ERROR,
                    file_type_code=file_type_codes.PACEMAKER_AUTHKEY,
                    file_path=LocalConfig.PCMK_AUTHKEY_PATH,
                    operation=RawFileError.ACTION_READ,
                )
            ],
        )

    def test_validate_addr_already_exists(self):
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
        )
        #more validation tests in pcs/lib/cib/test/test_resource_remote_node.py
        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(
                self.env_assist.get_env(),
                node_addr=NODE_1,
            ),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.ID_ALREADY_EXISTS,
                    id=NODE_1
                )
            ]
        )

    def test_unknown_host(self):
        self.config.env.set_known_hosts_dests({
            NODE_1: NODE_1_DEST_LIST,
            NODE_2: NODE_2_DEST_LIST,
        })
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(self.env_assist.get_env()),
            []
        )
        self.env_assist.assert_reports([
            fixture.error(
                report_codes.HOST_NOT_FOUND,
                force_code=report_codes.SKIP_OFFLINE_NODES,
                host_list=[NODE_NAME],
            )
        ])

    def test_unknown_host_skip_offline(self):
        pcmk_authkey_content = b"password"
        self.config.env.set_known_hosts_dests({
            NODE_1: NODE_1_DEST_LIST,
            NODE_2: NODE_2_DEST_LIST,
        })
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .local.authkey_exists(return_value=True)
            .local.open_authkey(pcmk_authkey_content)
            .env.push_cib(resources=FIXTURE_RESOURCES)
        )
        node_add_remote(self.env_assist.get_env(), skip_offline_nodes=True)
        self.env_assist.assert_reports(
            fixture_reports_new_node_unreachable(NODE_NAME)
        )

    @mock.patch("pcs.lib.commands.remote_node.generate_binary_key")
    def test_unknown_host_skip_offline_authkey_distribution(
        self, generate_binary_key
    ):
        generate_binary_key.return_value = b"password"
        self.config.env.set_known_hosts_dests({
            NODE_1: NODE_1_DEST_LIST,
            NODE_2: NODE_2_DEST_LIST,
        })
        (self.config
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .local.authkey_exists(return_value=False)
            .local.distribute_authkey(
                communication_list=[
                    dict(label=NODE_1, dest_list=NODE_1_DEST_LIST),
                    dict(label=NODE_2, dest_list=NODE_2_DEST_LIST),
                ],
                pcmk_authkey_content=generate_binary_key.return_value,
            )
            .env.push_cib(resources=FIXTURE_RESOURCES)
        )
        node_add_remote(self.env_assist.get_env(), skip_offline_nodes=True)
        generate_binary_key.assert_called_once_with(random_bytes_count=256)
        self.env_assist.assert_reports(
            fixture_reports_new_node_unreachable(NODE_NAME)
            + [
            fixture.info(
                report_codes.FILES_DISTRIBUTION_STARTED,
                file_list=["pacemaker authkey"],
                node_list=[NODE_1, NODE_2],
            ),
            fixture.info(
                report_codes.FILE_DISTRIBUTION_SUCCESS,
                file_description="pacemaker authkey",
                node=NODE_1,
            ),
            fixture.info(
                report_codes.FILE_DISTRIBUTION_SUCCESS,
                file_description="pacemaker authkey",
                node=NODE_2,
            ),
        ])

    def test_some_node_names_missing(self):
        self._config_success_base()
        (self.config
            .env.set_known_hosts_dests({
                "rh7-1": [Destination("rh7-1", 2224)],
                "rh7-2": [Destination("rh7-2", 2224)],
                NODE_NAME: NODE_DEST_LIST,
            })
            .corosync_conf.load(
                filename="corosync-some-node-names.conf",
                instead="corosync_conf.load"
            )
        )

        node_add_remote(self.env_assist.get_env())
        self.env_assist.assert_reports(
            REPORTS.warn(
                "missing_node_names_in_corosync",
                report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                fatal=False
            )
        )

    def test_all_node_names_missing(self):
        self._config_success_base()
        self.config.corosync_conf.load(
            filename="corosync-no-node-names.conf",
            instead="corosync_conf.load"
        )
        node_add_remote(self.env_assist.get_env())
        self.env_assist.assert_reports(
            REPORTS.warn(
                "missing_node_names_in_corosync",
                report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                fatal=False
            )
        )


class NotLive(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.env.set_known_hosts_dests(KNOWN_HOSTS_DESTS)
        with open(rc("cib-empty.xml")) as cib_file:
            self.config.env.set_cib_data(cib_file.read())

    def test_addr_specified(self):
        (self.config
            .runner.cib.load()
            .runner.pcmk.load_agent(agent_name="ocf:pacemaker:remote")
            .env.push_cib(resources=FIXTURE_RESOURCES)
        )
        node_add_remote(self.env_assist.get_env())
        self.env_assist.assert_reports(fixture_reports_not_live_cib(NODE_NAME))

    def test_addr_not_specified(self):
        (self.config
            .runner.cib.load()
            .runner.pcmk.load_agent(agent_name="ocf:pacemaker:remote")
            .env.push_cib(
                resources=FIXTURE_RESOURCES_TEMPLATE.format(
                    server=NODE_ADDR_PCSD
                )
            )
        )
        node_add_remote(self.env_assist.get_env(), no_node_addr=True)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    report_codes.USING_KNOWN_HOST_ADDRESS_FOR_HOST,
                    host_name=NODE_NAME,
                    address=NODE_ADDR_PCSD,
                ),
            ]
            +
            fixture_reports_not_live_cib(NODE_NAME)
        )

    def test_unknown_host_addr_not_specified(self):
        self.config.env.set_known_hosts_dests(dict())
        (self.config
            .runner.cib.load()
            .runner.pcmk.load_agent(agent_name="ocf:pacemaker:remote")
            .env.push_cib(
                resources=FIXTURE_RESOURCES_TEMPLATE.format(
                    server=NODE_NAME
                )
            )
        )
        node_add_remote(self.env_assist.get_env(), no_node_addr=True)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    report_codes.USING_KNOWN_HOST_ADDRESS_FOR_HOST,
                    host_name=NODE_NAME,
                    address=NODE_NAME,
                ),
            ]
            +
            fixture_reports_not_live_cib(NODE_NAME)
        )

    def test_unknown_host_addr_specified(self):
        self.config.env.set_known_hosts_dests(dict())
        (self.config
            .runner.cib.load()
            .runner.pcmk.load_agent(agent_name="ocf:pacemaker:remote")
            .env.push_cib(
                resources=FIXTURE_RESOURCES_TEMPLATE.format(
                    server="addr"
                )
            )
        )
        node_add_remote(self.env_assist.get_env(), node_addr="addr")
        self.env_assist.assert_reports(fixture_reports_not_live_cib(NODE_NAME))

    def test_wait(self):
        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(self.env_assist.get_env(), wait=1),
            [
                fixture.error(
                    report_codes.WAIT_FOR_IDLE_NOT_LIVE_CLUSTER,
                ),
            ],
            expected_in_processor=False
        )


class WithWait(TestCase):
    def setUp(self):
        self. wait = 1
        self.env_assist, self.config = get_env_tools(self)
        (self.config
            .env.set_known_hosts_dests(KNOWN_HOSTS_DESTS)
            .runner.pcmk.can_wait()
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
                ],
            )
            .local.get_host_info(NODE_NAME, NODE_DEST_LIST)
            .local.push_existing_authkey_to_remote(NODE_NAME, NODE_DEST_LIST)
            .local.run_pacemaker_remote(NODE_NAME, NODE_DEST_LIST)
            .env.push_cib(resources=FIXTURE_RESOURCES, wait=self.wait)
         )

    def test_success_when_resource_started(self):
        (self.config
            .runner.pcmk.load_state(raw_resources=dict(
                resource_id=NODE_NAME,
                resource_agent="ocf::pacemaker:remote",
                node_name=NODE_1,
            ))
            # self.config.fs is used to mock authkey file existence. Therefore,
            # filesystem mocking is active and we need to cover it. We tell the
            # file doesn't exist, because we aren't currently mocking the
            # function which reads it:
            #   if os.path.isfile(settings.crm_mon_schema):
            #       etree.RelaxNG(file=settings.crm_mon_schema).assertValid(dom)
            .fs.isfile(settings.crm_mon_schema, return_value=False)
        )
        node_add_remote(self.env_assist.get_env(), wait=self.wait)
        self.env_assist.assert_reports(
            REPORTS
                .info(
                    "resource_running",
                    report_codes.RESOURCE_RUNNING_ON_NODES,
                    roles_with_nodes={"Started": [NODE_1]},
                    resource_id=NODE_NAME
                )
        )

    def test_fail_when_resource_not_started(self):
        (self.config
            .runner.pcmk.load_state(raw_resources=dict(
                resource_id=NODE_NAME,
                resource_agent="ocf::pacemaker:remote",
                node_name=NODE_1,
                failed="true",
            ))
            # self.config.fs is used to mock authkey file existence. Therefore,
            # filesystem mocking is active and we need to cover it. We tell the
            # file doesn't exist, because we aren't currently mocking the
            # function which reads it:
            #   if os.path.isfile(settings.crm_mon_schema):
            #       etree.RelaxNG(file=settings.crm_mon_schema).assertValid(dom)
            .fs.isfile(settings.crm_mon_schema, return_value=False)
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(self.env_assist.get_env(), wait=self.wait),
            [
                fixture.error(
                    report_codes.RESOURCE_DOES_NOT_RUN,
                    resource_id=NODE_NAME,
                )
            ]
        )
        self.env_assist.assert_reports(REPORTS)


class AddRemotePcmkRemoteService(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        (self.config
            .env.set_known_hosts_dests(KNOWN_HOSTS_DESTS)
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
                ],
            )
            .local.get_host_info(NODE_NAME, NODE_DEST_LIST)
            .local.push_existing_authkey_to_remote(NODE_NAME, NODE_DEST_LIST)
        )

    def test_fails_when_offline(self):
        (self.config
            .local.run_pacemaker_remote(
                NODE_NAME, NODE_DEST_LIST, **FAIL_HTTP_KWARGS
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(self.env_assist.get_env())
        )

        self.env_assist.assert_reports(
            REPORTS[:"pcmk_remote_enable_success"]
            +
            EXTRA_REPORTS.select("manage_services_connection_failed")
        )

    def test_fail_when_remotely_fail(self):
        (self.config
            .local.run_pacemaker_remote(NODE_NAME, NODE_DEST_LIST, result={
                "code": "fail",
                "message": "Action failed",
            })
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(self.env_assist.get_env())
        )
        self.env_assist.assert_reports(
            REPORTS[:"pcmk_remote_enable_success"] + EXTRA_REPORTS.select(
                "pcmk_remote_enable_failed",
                "pcmk_remote_start_failed",
            )
        )

    def test_forceable_when_remotely_fail(self):
        (self.config
            .local.run_pacemaker_remote(NODE_NAME, NODE_DEST_LIST, result={
                "code": "fail",
                "message": "Action failed",
            })
            .env.push_cib(resources=FIXTURE_RESOURCES)
        )
        node_add_remote(
            self.env_assist.get_env(),
            allow_pacemaker_remote_service_fail=True
        )

        self.env_assist.assert_reports(
            REPORTS[:"pcmk_remote_enable_success"] + EXTRA_REPORTS.select(
                "pcmk_remote_enable_failed_warn",
                "pcmk_remote_start_failed_warn",
            )
        )


class AddRemoteAuthkeyDistribution(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        (self.config
            .env.set_known_hosts_dests(KNOWN_HOSTS_DESTS)
            .local.load_cluster_configs(cluster_node_list=[NODE_1, NODE_2])
            .http.host.check_auth(
                communication_list=[
                    dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
                ],
            )
            .local.get_host_info(NODE_NAME, NODE_DEST_LIST)
        )

    def test_fails_when_offline(self):
        pcmk_authkey_content = b"password"
        (self.config
            .local.authkey_exists(return_value=True)
            .local.open_authkey(pcmk_authkey_content)
            .local.distribute_authkey(
                communication_list=[
                    dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
                ],
                pcmk_authkey_content=pcmk_authkey_content,
                **FAIL_HTTP_KWARGS
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(self.env_assist.get_env())
        )
        self.env_assist.assert_reports(
            REPORTS[:"authkey_distribution_success"]
            +
            EXTRA_REPORTS.only(
                "manage_services_connection_failed",
                command="remote/put_file",
            )
        )

    def test_fail_when_remotely_fail(self):
        (self.config
            .local.push_existing_authkey_to_remote(
                NODE_NAME,
                NODE_DEST_LIST,
                distribution_result={
                    "code": "conflict",
                    "message": "",
                }
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: node_add_remote(self.env_assist.get_env())
        )

        self.env_assist.assert_reports(
            REPORTS[:"authkey_distribution_success"]
            +
            EXTRA_REPORTS.select("authkey_distribution_failed")
        )

    def test_forceable_when_remotely_fail(self):
        (self.config
            .local.push_existing_authkey_to_remote(
                NODE_NAME,
                NODE_DEST_LIST,
                distribution_result={
                    "code": "conflict",
                    "message": "",
                }
            )
            .local.run_pacemaker_remote(NODE_NAME, NODE_DEST_LIST)
            .env.push_cib(resources=FIXTURE_RESOURCES)
        )

        node_add_remote(
            self.env_assist.get_env(),
            allow_incomplete_distribution=True,
        )

        self.env_assist.assert_reports(
            REPORTS.remove("authkey_distribution_success")
            +
            EXTRA_REPORTS.select("authkey_distribution_failed_warn")
        )



from unittest import mock, TestCase
from lxml import etree

from pcs_test.tools.assertions import assert_report_item_list_equal

from pcs.common import report_codes
from pcs.lib import resource_agent as lib_ra
from pcs.lib.errors import ReportItemSeverity as severity
from pcs.lib.external import CommandRunner


class ValidateParameters(TestCase):
    def setUp(self):
        self.agent = lib_ra.StonithAgent(
            mock.MagicMock(spec_set=CommandRunner),
            "fence_dummy"
        )
        self.metadata = etree.XML("""
            <resource-agent>
                <parameters>
                    <parameter name="test_param" required="0">
                        <longdesc>Long description</longdesc>
                        <shortdesc>short description</shortdesc>
                        <content type="string" default="default_value" />
                    </parameter>
                    <parameter name="required_param" required="1">
                        <content type="boolean" />
                    </parameter>
                    <parameter name="action">
                        <content type="string" default="reboot" />
                        <shortdesc>Fencing action</shortdesc>
                    </parameter>
                </parameters>
            </resource-agent>
        """)
        patcher = mock.patch.object(lib_ra.StonithAgent, "_get_metadata")
        self.addCleanup(patcher.stop)
        self.get_metadata = patcher.start()
        self.get_metadata.return_value = self.metadata

        patcher_fenced = mock.patch.object(
            lib_ra.FencedMetadata, "_get_metadata"
        )
        self.addCleanup(patcher_fenced.stop)
        self.get_fenced_metadata = patcher_fenced.start()
        self.get_fenced_metadata.return_value = etree.XML("""
            <resource-agent>
                <parameters />
            </resource-agent>
        """)
        self.report_error = (
            severity.ERROR,
            report_codes.DEPRECATED_OPTION,
            {
                "option_name": "action",
                "option_type": "stonith",
                "replaced_by": [
                    "pcmk_off_action",
                    "pcmk_reboot_action"
                ],
            },
            report_codes.FORCE_OPTIONS
        )
        self.report_warning = (
            severity.WARNING,
            report_codes.DEPRECATED_OPTION,
            {
                "option_name": "action",
                "option_type": "stonith",
                "replaced_by": [
                    "pcmk_off_action",
                    "pcmk_reboot_action"
                ],
            },
            None
        )


class ValidateParametersCreate(ValidateParameters):
    def test_action_is_deprecated(self):
        assert_report_item_list_equal(
            self.agent.validate_parameters_create({
                "action": "reboot",
                "required_param": "value",
            }),
            [
                self.report_error,
            ],
        )

    def test_action_is_deprecated_forced(self):
        assert_report_item_list_equal(
            self.agent.validate_parameters_create({
                "action": "reboot",
                "required_param": "value",
            }, force=True),
            [
                self.report_warning,
            ],
        )

    def test_action_not_reported_deprecated_when_empty(self):
        assert_report_item_list_equal(
            self.agent.validate_parameters_create({
                "action": "",
                "required_param": "value",
            }),
            [
            ],
        )


class ValidateParametersUpdate(ValidateParameters):
    def test_action_is_deprecated(self):
        assert_report_item_list_equal(
            self.agent.validate_parameters_update(
                {
                    "required_param": "value",
                },
                {
                    "action": "reboot",
                }
            ),
            [
                self.report_error,
            ],
        )

    def test_action_not_reported_when_not_updated(self):
        assert_report_item_list_equal(
            self.agent.validate_parameters_update(
                {
                    "required_param": "value",
                    "action": "reboot",
                },
                {
                    "required_param": "value2",
                }
            ),
            [
            ],
        )

    def test_action_is_deprecated_when_set_already(self):
        assert_report_item_list_equal(
            self.agent.validate_parameters_update(
                {
                    "required_param": "value",
                    "action": "off",
                },
                {
                    "action": "reboot",
                }
            ),
            [
                self.report_error,
            ],
        )

    def test_action_is_deprecated_forced(self):
        assert_report_item_list_equal(
            self.agent.validate_parameters_update(
                {
                    "required_param": "value",
                },
                {
                    "action": "reboot",
                },
                force=True
            ),
            [
                self.report_warning,
            ],
        )

    def test_action_not_reported_deprecated_when_empty(self):
        assert_report_item_list_equal(
            self.agent.validate_parameters_update(
                {
                    "required_param": "value",
                    "action": "reboot",
                },
                {
                    "action": "",
                },
            ),
            [
            ],
        )


@mock.patch.object(lib_ra.StonithAgent, "get_actions")
class StonithAgentMetadataGetCibDefaultActions(TestCase):
    fixture_actions = [
        {"name": "custom1", "timeout": "40s"},
        {"name": "custom2", "interval": "25s", "timeout": "60s"},
        {"name": "meta-data"},
        {"name": "monitor", "interval": "10s", "timeout": "30s"},
        {"name": "start", "interval": "40s"},
        {"name": "status", "interval": "15s", "timeout": "20s"},
        {"name": "validate-all"},
    ]

    def setUp(self):
        self.agent = lib_ra.StonithAgent(
            mock.MagicMock(spec_set=CommandRunner),
            "fence_dummy"
        )

    def test_select_only_actions_for_cib(self, get_actions):
        get_actions.return_value = self.fixture_actions
        self.assertEqual(
            [
                {"name": "monitor", "interval": "10s", "timeout": "30s"}
            ],
            self.agent.get_cib_default_actions()
        )

    def test_select_only_necessary_actions_for_cib(self, get_actions):
        get_actions.return_value = self.fixture_actions
        self.assertEqual(
            [
                {"name": "monitor", "interval": "10s", "timeout": "30s"}
            ],
            self.agent.get_cib_default_actions(necessary_only=True)
        )



from pcs import (
    pcsd,
    usage,
)
from pcs.cli.common.errors import raise_command_replaced
from pcs.cli.common.routing import create_router


pcsd_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: usage.pcsd(argv),
        "deauth": pcsd.pcsd_deauth,
        "certkey": pcsd.pcsd_certkey,
        "sync-certificates": pcsd.pcsd_sync_certs,

        # removed commands
        # These print error messages which point users to the changes section in
        # pcs manpage.
        # To be removed in the next significant version.
        "clear-auth": lambda lib, argv, modifiers: raise_command_replaced(
            "pcs host deauth", "pcs pcsd deauth"
        ),
    },
    ["pcsd"]
)



import logging
import unittest

from pcs.cli.common.reports import (
    LibraryReportProcessorToConsole as ReportProcessor
)
from pcs.lib.external import CommandRunner
from pcs.lib.resource_agent import StonithAgent


def __can_load_xvm_fence_agent():
    # pylint: disable=bare-except
    try:
        runner = CommandRunner(logging.getLogger("test"), ReportProcessor())
        StonithAgent(runner, "fence_xvm").validate_metadata()
        return True
    except:
        return False


need_load_xvm_fence_agent = unittest.skipUnless(
    __can_load_xvm_fence_agent(),
    "test requires the successful load of 'fence_xvm' agent"
)



import json
import re
from textwrap import dedent
from unittest import mock, TestCase

from pcs.common import report_codes
from pcs.lib.corosync.config_facade import ConfigFacade as CorosyncConfigFacade
from pcs.lib.corosync.config_parser import Section as CorosyncSection

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_raise_library_error
from pcs_test.tools.command_env import get_env_tools

class PushCorosyncConfLiveBase(TestCase):
    def setUp(self):
        self.env_assistant, self.config = get_env_tools(self)
        self.corosync_conf_facade = None
        self.node_labels = ["node-1", "node-2"]
        self.config.env.set_known_nodes(self.node_labels)
        self.corosync_conf_text = "corosync conf"
        self.fixture_corosync_conf()

    def fixture_corosync_conf(self, node1_name=True, node2_name=True):
        # nodelist is enough, nothing else matters for the tests
        config = dedent("""\
            nodelist {
                node {
                    ring0_addr: node-1-addr
                    nodeid: 1
                    name: node-1
                }

                node {
                    ring0_addr: node-2-addr
                    nodeid: 2
                    name: node-2
                }
            }
            """
        )
        if not node1_name:
            config = re.sub(r"\s+name: node-1\n", "\n", config)
        if not node2_name:
            config = re.sub(r"\s+name: node-2\n", "\n", config)
        self.corosync_conf_text = config
        self.corosync_conf_facade = CorosyncConfigFacade.from_string(config)
        CorosyncConfigFacade.need_stopped_cluster = False
        CorosyncConfigFacade.need_qdevice_reload = False


@mock.patch("pcs.lib.external.is_systemctl", lambda: True)
class PushCorosyncConfLiveNoQdeviceTest(PushCorosyncConfLiveBase):
    def test_some_node_names_missing(self):
        self.fixture_corosync_conf(node1_name=False)

        self.env_assistant.assert_raise_library_error(
            lambda: self.env_assistant.get_env().push_corosync_conf(
                self.corosync_conf_facade
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=True,
                ),
            ]
        )

    def test_all_node_names_missing(self):
        self.fixture_corosync_conf(node1_name=False, node2_name=False)

        self.env_assistant.assert_raise_library_error(
            lambda: self.env_assistant.get_env().push_corosync_conf(
                self.corosync_conf_facade
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=True,
                ),
            ]
        )

    def test_dont_need_stopped_cluster(self):
        (self.config
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                node_labels=self.node_labels
            )
            .http.corosync.reload_corosync_conf(
                node_labels=self.node_labels[:1]
            )
        )
        self.env_assistant.get_env().push_corosync_conf(
            self.corosync_conf_facade
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-2",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_RELOADED,
                node="node-1"
            ),
        ])

    def test_dont_need_stopped_cluster_error(self):
        (self.config
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                communication_list=[
                    {
                        "label": "node-1",
                    },
                    {
                        "label": "node-2",
                        "response_code": 400,
                        "output": "Failed"
                    },
                ]
            )
        )
        env = self.env_assistant.get_env()
        self.env_assistant.assert_raise_library_error(
            lambda: env.push_corosync_conf(self.corosync_conf_facade),
            []
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-1",
            ),
            fixture.error(
                report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                force_code=report_codes.SKIP_OFFLINE_NODES,
                node="node-2",
                command="remote/set_corosync_conf",
                reason="Failed",
            ),
            fixture.error(
                report_codes.COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR,
                force_code=report_codes.SKIP_OFFLINE_NODES,
                node="node-2",
            ),
        ])

    def test_dont_need_stopped_cluster_error_skip_offline(self):
        (self.config
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                communication_list=[
                    {
                        "label": "node-1",
                        "response_code": 400,
                        "output": "Failed"
                    },
                    {
                        "label": "node-2",
                    },
                ]
            )
            .http.corosync.reload_corosync_conf(
                communication_list=[
                    [
                        {
                            "label": self.node_labels[0],
                            "response_code": 400,
                            "output": "Failed"
                        },
                    ],
                    [
                        {
                            "label": self.node_labels[1],
                        },
                    ],
                ]
            )
        )
        self.env_assistant.get_env().push_corosync_conf(
            self.corosync_conf_facade, skip_offline_nodes=True
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.warn(
                report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                node="node-1",
                command="remote/set_corosync_conf",
                reason="Failed",
            ),
            fixture.warn(
                report_codes.COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-2",
            ),
            fixture.warn(
                report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                node="node-1",
                command="remote/reload_corosync_conf",
                reason="Failed",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_RELOADED,
                node="node-2"
            ),
        ])

    def test_reload_on_another_node(self):
        (self.config
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                node_labels=self.node_labels
            )
            .http.corosync.reload_corosync_conf(
                communication_list=[
                    [
                        {
                            "label": self.node_labels[0],
                            "response_code": 200,
                            "output": json.dumps(
                                dict(code="not_running", message="not running")
                            ),
                        },
                    ],
                    [
                        {
                            "label": self.node_labels[1],
                        },
                    ],
                ]
            )
        )
        self.env_assistant.get_env().push_corosync_conf(
            self.corosync_conf_facade
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-2",
            ),
            fixture.warn(
                report_codes.COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE,
                node="node-1"
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_RELOADED,
                node="node-2"
            ),
        ])

    def test_reload_not_successful(self):
        (self.config
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                node_labels=self.node_labels
            )
            .http.corosync.reload_corosync_conf(
                communication_list=[
                    [
                        {
                            "label": self.node_labels[0],
                            "response_code": 200,
                            "output": json.dumps(
                                dict(code="not_running", message="not running")
                            ),
                        },
                    ],
                    [
                        {
                            "label": self.node_labels[1],
                            "response_code": 200,
                            "output": "not a json",
                        },
                    ],
                ]
            )
        )
        self.env_assistant.assert_raise_library_error(
            lambda: self.env_assistant.get_env().push_corosync_conf(
                self.corosync_conf_facade
            ),
            []
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-2",
            ),
            fixture.warn(
                report_codes.COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE,
                node="node-1"
            ),
            fixture.warn(
                report_codes.INVALID_RESPONSE_FORMAT,
                node="node-2"
            ),
            fixture.error(
                report_codes.UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE
            ),
        ])

    def test_reload_corosync_not_running_anywhere(self):
        (self.config
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                node_labels=self.node_labels
            )
            .http.corosync.reload_corosync_conf(
                communication_list=[
                    [
                        {
                            "label": node,
                            "response_code": 200,
                            "output": json.dumps(
                                dict(code="not_running", message="not running")
                            ),
                        },
                    ] for node in self.node_labels
                ]
            )
        )
        self.env_assistant.get_env().push_corosync_conf(
            self.corosync_conf_facade
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-2",
            ),
            fixture.warn(
                report_codes.COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE,
                node="node-1"
            ),
            fixture.warn(
                report_codes.COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE,
                node="node-2"
            ),
        ])

    def test_need_stopped_cluster(self):
        self.corosync_conf_facade.need_stopped_cluster = True
        (self.config
            .http.corosync.check_corosync_offline(
                node_labels=self.node_labels
            )
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                node_labels=self.node_labels
            )
        )
        self.env_assistant.get_env().push_corosync_conf(
            self.corosync_conf_facade
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
            fixture.info(
                report_codes.COROSYNC_NOT_RUNNING_ON_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_NOT_RUNNING_ON_NODE,
                node="node-2",
            ),
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-2",
            ),
        ])

    def test_need_stopped_cluster_not_stopped(self):
        self.corosync_conf_facade.need_stopped_cluster = True
        (self.config
            .http.corosync.check_corosync_offline(
                communication_list=[
                    {
                        "label": self.node_labels[0],
                        "output": '{"corosync":true}'
                    }
                ]
                +
                [
                    {
                        "label": node,
                    }
                    for node in self.node_labels[1:]
                ]
            )
        )
        env = self.env_assistant.get_env()
        self.env_assistant.assert_raise_library_error(
            lambda: env.push_corosync_conf(self.corosync_conf_facade),
            []
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
            fixture.error(
                report_codes.COROSYNC_RUNNING_ON_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_NOT_RUNNING_ON_NODE,
                node="node-2",
            ),
        ])

    def test_need_stopped_cluster_not_stopped_skip_offline(self):
        # If we know for sure that corosync is running, skip_offline doesn't
        # matter.
        self.corosync_conf_facade.need_stopped_cluster = True
        (self.config
            .http.corosync.check_corosync_offline(
                communication_list=[
                    dict(
                        label="node-1",
                        output='{"corosync":true}',
                    ),
                    dict(
                        label="node-2",
                    ),
                ]
            )
        )
        env = self.env_assistant.get_env()
        self.env_assistant.assert_raise_library_error(
            lambda: env.push_corosync_conf(
                self.corosync_conf_facade, skip_offline_nodes=True
            ),
            []
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
            fixture.error(
                report_codes.COROSYNC_RUNNING_ON_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_NOT_RUNNING_ON_NODE,
                node="node-2",
            )
        ])

    def test_need_stopped_cluster_json_error(self):
        self.corosync_conf_facade.need_stopped_cluster = True
        (self.config
            .http.corosync.check_corosync_offline(
                communication_list=[
                    dict(
                        label="node-1",
                        output="{" # not valid json
                    ),
                    dict(
                        label="node-2",
                        # The expected key (/corosync) is missing, we don't
                        # care about version 2 status key
                        # (/services/corosync/running)
                        output='{"services":{"corosync":{"running":true}}}'
                    ),
                ]
            )
        )
        env = self.env_assistant.get_env()
        self.env_assistant.assert_raise_library_error(
            lambda: env.push_corosync_conf(self.corosync_conf_facade),
            []
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
            fixture.error(
                report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                force_code=report_codes.SKIP_OFFLINE_NODES,
                node="node-1",
            ),
            fixture.error(
                report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                force_code=report_codes.SKIP_OFFLINE_NODES,
                node="node-2",
            ),
        ])

    def test_need_stopped_cluster_comunnication_failure(self):
        self.corosync_conf_facade.need_stopped_cluster = True
        (self.config
            .http.corosync.check_corosync_offline(
                communication_list=[
                    dict(
                        label="node-1",
                    ),
                    dict(
                        label="node-2",
                        response_code=401,
                        output="""{"notauthorized":"true"}"""
                    ),
                ]
            )
        )
        env = self.env_assistant.get_env()
        self.env_assistant.assert_raise_library_error(
            lambda: env.push_corosync_conf(self.corosync_conf_facade),
            []
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
            fixture.info(
                report_codes.COROSYNC_NOT_RUNNING_ON_NODE,
                node="node-1",
            ),
            fixture.error(
                report_codes.NODE_COMMUNICATION_ERROR_NOT_AUTHORIZED,
                force_code=report_codes.SKIP_OFFLINE_NODES,
                node="node-2",
            ),
            fixture.error(
                report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                force_code=report_codes.SKIP_OFFLINE_NODES,
                node="node-2",
            ),
        ])

    def test_need_stopped_cluster_comunnication_failures_skip_offline(self):
        # If we don't know if corosync is running, skip_offline matters.
        self.corosync_conf_facade.need_stopped_cluster = True
        (self.config
            .http.corosync.check_corosync_offline(
                communication_list=[
                    dict(
                        label="node-1",
                        response_code=401,
                        output="""{"notauthorized":"true"}"""
                    ),
                    dict(
                        label="node-2",
                        output="{" # not valid json
                    ),
                ]
            )
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                communication_list=[
                    dict(
                        label="node-1",
                        response_code=401,
                        output="""{"notauthorized":"true"}""",
                    ),
                    dict(
                        label="node-2",
                    )
                ]
            )
        )
        self.env_assistant.get_env().push_corosync_conf(
            self.corosync_conf_facade, skip_offline_nodes=True
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_NOT_RUNNING_CHECK_STARTED),
            fixture.warn(
                report_codes.NODE_COMMUNICATION_ERROR_NOT_AUTHORIZED,
                node="node-1",
                reason="HTTP error: 401",
                command="remote/status",
            ),
            fixture.warn(
                report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                node="node-1",
            ),
            fixture.warn(
                report_codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR,
                node="node-2",
            ),
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.warn(
                report_codes.NODE_COMMUNICATION_ERROR_NOT_AUTHORIZED,
                node="node-1",
                reason="HTTP error: 401",
                command="remote/set_corosync_conf",
            ),
            fixture.warn(
                report_codes.COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-2",
            ),
        ])


@mock.patch("pcs.lib.external.is_systemctl", lambda: True)
class PushCorosyncConfLiveWithQdeviceTest(PushCorosyncConfLiveBase):
    def test_some_node_names_missing(self):
        self.fixture_corosync_conf(node1_name=False)
        self.corosync_conf_facade.need_qdevice_reload = True

        self.env_assistant.assert_raise_library_error(
            lambda: self.env_assistant.get_env().push_corosync_conf(
                self.corosync_conf_facade
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=True,
                ),
            ]
        )

    def test_all_node_names_missing(self):
        self.fixture_corosync_conf(node1_name=False, node2_name=False)
        self.corosync_conf_facade.need_qdevice_reload = True

        self.env_assistant.assert_raise_library_error(
            lambda: self.env_assistant.get_env().push_corosync_conf(
                self.corosync_conf_facade
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=True,
                ),
            ]
        )

    def test_qdevice_reload(self):
        self.corosync_conf_facade.need_qdevice_reload = True
        (self.config
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                node_labels=self.node_labels
            )
            .http.corosync.reload_corosync_conf(
                node_labels=self.node_labels[:1]
            )
            .http.corosync.qdevice_client_stop(
                node_labels=self.node_labels
            )
            .http.corosync.qdevice_client_start(
                node_labels=self.node_labels
            )
        )

        self.env_assistant.get_env().push_corosync_conf(
            self.corosync_conf_facade
        )

        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-2",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_RELOADED,
                node="node-1"
            ),
            fixture.info(report_codes.QDEVICE_CLIENT_RELOAD_STARTED),
            fixture.info(
                report_codes.SERVICE_STOP_SUCCESS,
                node="node-1",
                service="corosync-qdevice",
                instance=None,
            ),
            fixture.info(
                report_codes.SERVICE_STOP_SUCCESS,
                node="node-2",
                service="corosync-qdevice",
                instance=None,
            ),
            fixture.info(
                report_codes.SERVICE_START_SUCCESS,
                node="node-1",
                service="corosync-qdevice",
                instance=None,
            ),
            fixture.info(
                report_codes.SERVICE_START_SUCCESS,
                node="node-2",
                service="corosync-qdevice",
                instance=None,
            ),
        ])

    def test_qdevice_reload_corosync_stopped(self):
        self.corosync_conf_facade.need_qdevice_reload = True
        (self.config
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                node_labels=self.node_labels
            )
            .http.corosync.reload_corosync_conf(
                communication_list=[
                    [
                        {
                            "label": label,
                            "response_code": 200,
                            "output": json.dumps(
                                dict(code="not_running", message="")
                            ),
                        },
                    ] for label in self.node_labels
                ]
            )
            .http.corosync.qdevice_client_stop(
                node_labels=self.node_labels
            )
            .http.corosync.qdevice_client_start(
                communication_list=[
                    {
                        "label": label,
                        "output": "corosync is not running, skipping",
                    }
                    for label in self.node_labels
                ]
            )
        )

        self.env_assistant.get_env().push_corosync_conf(
            self.corosync_conf_facade
        )

        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-2",
            ),
            fixture.warn(
                report_codes.COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE,
                node="node-1",
            ),
            fixture.warn(
                report_codes.COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE,
                node="node-2",
            ),
            fixture.info(report_codes.QDEVICE_CLIENT_RELOAD_STARTED),
            fixture.info(
                report_codes.SERVICE_STOP_SUCCESS,
                node="node-1",
                service="corosync-qdevice",
                instance=None,
            ),
            fixture.info(
                report_codes.SERVICE_STOP_SUCCESS,
                node="node-2",
                service="corosync-qdevice",
                instance=None,
            ),
            fixture.info(
                report_codes.SERVICE_START_SKIPPED,
                node="node-1",
                service="corosync-qdevice",
                instance=None,
                reason="corosync is not running",
            ),
            fixture.info(
                report_codes.SERVICE_START_SKIPPED,
                node="node-2",
                service="corosync-qdevice",
                instance=None,
                reason="corosync is not running",
            ),
        ])

    def test_qdevice_reload_failures(self):
        # This also tests that failing to stop qdevice on a node doesn't prevent
        # starting qdevice on the same node.
        self.corosync_conf_facade.need_qdevice_reload = True
        (self.config
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                node_labels=self.node_labels
            )
            .http.corosync.reload_corosync_conf(
                node_labels=self.node_labels[:1]
            )
            .http.corosync.qdevice_client_stop(
                communication_list=[
                    dict(
                        label="node-1",
                    ),
                    dict(
                        label="node-2",
                        response_code=400,
                        output="error",
                    ),
                ]
            )
            .http.corosync.qdevice_client_start(
                communication_list=[
                    dict(
                        label="node-1",
                        errno=8,
                        error_msg="failure",
                        was_connected=False,
                    ),
                    dict(
                        label="node-2",
                    ),
                ]
            )
        )

        env = self.env_assistant.get_env()
        self.env_assistant.assert_raise_library_error(
            lambda: env.push_corosync_conf(self.corosync_conf_facade),
            []
        )

        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-2",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_RELOADED,
                node="node-1"
            ),
            fixture.info(report_codes.QDEVICE_CLIENT_RELOAD_STARTED),
            fixture.info(
                report_codes.SERVICE_STOP_SUCCESS,
                node="node-1",
                service="corosync-qdevice",
                instance=None,
            ),
            fixture.error(
                report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                force_code=report_codes.SKIP_OFFLINE_NODES,
                node="node-2",
                reason="error",
                command="remote/qdevice_client_stop",
            ),
            fixture.error(
                report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                force_code=report_codes.SKIP_OFFLINE_NODES,
                node="node-1",
                reason="failure",
                command="remote/qdevice_client_start",
            ),
            fixture.info(
                report_codes.SERVICE_START_SUCCESS,
                node="node-2",
                service="corosync-qdevice",
                instance=None,
            ),
        ])

    def test_qdevice_reload_failures_skip_offline(self):
        self.corosync_conf_facade.need_qdevice_reload = True
        (self.config
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                communication_list=[
                    dict(
                        label="node-1",
                    ),
                    dict(
                        label="node-2",
                        errno=8,
                        error_msg="failure",
                        was_connected=False,
                    ),
                ]
            )
            .http.corosync.reload_corosync_conf(
                communication_list=[
                    [
                        {
                            "label": self.node_labels[0],
                            "response_code": 400,
                            "output": "Failed"
                        },
                    ],
                    [
                        {
                            "label": self.node_labels[1],
                        },
                    ],
                ]
            )
            .http.corosync.qdevice_client_stop(
                communication_list=[
                    dict(
                        label="node-1",
                    ),
                    dict(
                        label="node-2",
                        response_code=400,
                        output="error",
                    ),
                ]
            )
            .http.corosync.qdevice_client_start(
                communication_list=[
                    dict(
                        label="node-1",
                        errno=8,
                        error_msg="failure",
                        was_connected=False,
                    ),
                    dict(
                        label="node-2",
                    ),
                ]
            )
        )

        env = self.env_assistant.get_env()
        env.push_corosync_conf(
            self.corosync_conf_facade, skip_offline_nodes=True
        )

        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-1",
            ),
            fixture.warn(
                report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                node="node-2",
                reason="failure",
                command="remote/set_corosync_conf",
            ),
            fixture.warn(
                report_codes.COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR,
                node="node-2",
            ),
            fixture.warn(
                report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                node="node-1",
                command="remote/reload_corosync_conf",
                reason="Failed",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_RELOADED,
                node="node-2"
            ),
            fixture.info(report_codes.QDEVICE_CLIENT_RELOAD_STARTED),
            fixture.info(
                report_codes.SERVICE_STOP_SUCCESS,
                node="node-1",
                service="corosync-qdevice",
                instance=None,
            ),
            fixture.warn(
                report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                node="node-2",
                reason="error",
                command="remote/qdevice_client_stop",
            ),
            fixture.warn(
                report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                node="node-1",
                reason="failure",
                command="remote/qdevice_client_start",
            ),
            fixture.info(
                report_codes.SERVICE_START_SUCCESS,
                node="node-2",
                service="corosync-qdevice",
                instance=None,
            ),
        ])

    def test_reload_not_successful(self):
        self.corosync_conf_facade.need_qdevice_reload = True
        (self.config
            .http.corosync.set_corosync_conf(
                self.corosync_conf_text,
                node_labels=self.node_labels
            )
            .http.corosync.reload_corosync_conf(
                communication_list=[
                    [
                        {
                            "label": self.node_labels[0],
                            "response_code": 200,
                            "output": json.dumps(
                                dict(code="not_running", message="not running")
                            ),
                        },
                    ],
                    [
                        {
                            "label": self.node_labels[1],
                            "response_code": 200,
                            "output": "not a json",
                        },
                    ],
                ]
            )
        )
        self.env_assistant.assert_raise_library_error(
            lambda: self.env_assistant.get_env().push_corosync_conf(
                self.corosync_conf_facade
            ),
            []
        )
        self.env_assistant.assert_reports([
            fixture.info(report_codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-1",
            ),
            fixture.info(
                report_codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE,
                node="node-2",
            ),
            fixture.warn(
                report_codes.COROSYNC_CONFIG_RELOAD_NOT_POSSIBLE,
                node="node-1"
            ),
            fixture.warn(
                report_codes.INVALID_RESPONSE_FORMAT,
                node="node-2"
            ),
            fixture.error(
                report_codes.UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE
            ),
        ])


class PushCorosyncConfFile(TestCase):
    def setUp(self):
        self.env_assistant, self.config = get_env_tools(test_case=self)
        self.config.env.set_corosync_conf_data("totem {\n    version: 2\n}\n")

    def test_success(self):
        new_corosync_conf_data = "totem {\n    version: 3\n}\n"
        self.config.env.push_corosync_conf(
            corosync_conf_text=new_corosync_conf_data
        )
        env = self.env_assistant.get_env()
        env.push_corosync_conf(
            CorosyncConfigFacade.from_string(new_corosync_conf_data)
        )


class PushCorosyncConfBadCharsMixin():
    def test_bad_section(self):
        section = CorosyncSection("sec#tion")
        section.add_attribute("name", "value")
        new_conf = CorosyncConfigFacade(section)

        env = self.env_assistant.get_env()
        assert_raise_library_error(
            lambda: env.push_corosync_conf(new_conf),
            fixture.error(
                report_codes.COROSYNC_CONFIG_CANNOT_SAVE_INVALID_NAMES_VALUES,
                section_name_list=["sec#tion"],
                attribute_name_list=[],
                attribute_value_pairs=[],
            )
        )

    def test_bad_attr(self):
        section = CorosyncSection("section")
        section.add_attribute("na#me", "value")
        new_conf = CorosyncConfigFacade(section)

        env = self.env_assistant.get_env()
        assert_raise_library_error(
            lambda: env.push_corosync_conf(new_conf),
            fixture.error(
                report_codes.COROSYNC_CONFIG_CANNOT_SAVE_INVALID_NAMES_VALUES,
                section_name_list=[],
                attribute_name_list=["section.na#me"],
                attribute_value_pairs=[],
            )
        )

    def test_bad_value(self):
        section = CorosyncSection("section")
        section.add_attribute("name", "va}l{ue")
        new_conf = CorosyncConfigFacade(section)

        env = self.env_assistant.get_env()
        assert_raise_library_error(
            lambda: env.push_corosync_conf(new_conf),
            fixture.error(
                report_codes.COROSYNC_CONFIG_CANNOT_SAVE_INVALID_NAMES_VALUES,
                section_name_list=[],
                attribute_name_list=[],
                attribute_value_pairs=[("section.name", "va}l{ue")],
            )
        )

    def test_bad_all(self):
        section = CorosyncSection("sec#tion")
        section.add_attribute("na#me", "va}l{ue")
        new_conf = CorosyncConfigFacade(section)

        env = self.env_assistant.get_env()
        assert_raise_library_error(
            lambda: env.push_corosync_conf(new_conf),
            fixture.error(
                report_codes.COROSYNC_CONFIG_CANNOT_SAVE_INVALID_NAMES_VALUES,
                section_name_list=["sec#tion"],
                attribute_name_list=["sec#tion.na#me"],
                attribute_value_pairs=[("sec#tion.na#me", "va}l{ue")],
            )
        )


class PushCorosyncConfBadCharsLive(PushCorosyncConfBadCharsMixin, TestCase):
    def setUp(self):
        self.env_assistant, self.config = get_env_tools(self)


class PushCorosyncConfBadCharsFile(PushCorosyncConfBadCharsMixin, TestCase):
    def setUp(self):
        self.env_assistant, self.config = get_env_tools(self)
        self.config.env.set_corosync_conf_data("totem {\n    version: 2\n}\n")


class GetCorosyncConfFile(TestCase):
    def setUp(self):
        self.corosync_conf_data = "totem {\n    version: 2\n}\n"
        self.env_assistant, self.config = get_env_tools(test_case=self)
        self.config.env.set_corosync_conf_data(self.corosync_conf_data)

    def test_success(self):
        env = self.env_assistant.get_env()
        self.assertFalse(env.is_corosync_conf_live)
        self.assertEqual(self.corosync_conf_data, env.get_corosync_conf_data())
        self.assertEqual(
            self.corosync_conf_data, env.get_corosync_conf().config.export()
        )


class GetCorosyncConfLive(TestCase):
    def setUp(self):
        self.env_assistant, self.config = get_env_tools(self)

    def test_success(self):
        corosync_conf_data = "totem {\n    version: 2\n}\n"
        self.config.corosync_conf.load_content(corosync_conf_data)
        env = self.env_assistant.get_env()
        self.assertTrue(env.is_corosync_conf_live)
        self.assertEqual(
            corosync_conf_data, env.get_corosync_conf().config.export()
        )



from unittest import mock, TestCase

from lxml import etree

from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_report_item_equal,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.xml import get_xml_manipulation_creator_from_file

from pcs.common import report_codes
from pcs.lib.pacemaker import state
from pcs.lib.pacemaker.state import (
    ClusterState,
    _Attrs,
    _Children,
)
from pcs.lib.errors import ReportItemSeverity as severities

# pylint: disable=no-self-use, protected-access

class AttrsTest(TestCase):
    def test_get_declared_attr(self):
        attrs = _Attrs('test', {'node-name': 'node1'}, {'name': 'node-name'})
        self.assertEqual('node1', attrs.name)

    def test_raises_on_undeclared_attribute(self):
        attrs = _Attrs('test', {'node-name': 'node1'}, {})
        self.assertRaises(AttributeError, lambda: attrs.name)

    def test_raises_on_missing_required_attribute(self):
        attrs = _Attrs('test', {}, {'name': 'node-name'})
        self.assertRaises(AttributeError, lambda: attrs.name)

    def test_attr_transformation_success(self):
        attrs = _Attrs('test', {'number': '7'}, {'count': ('number', int)})
        self.assertEqual(7, attrs.count)

    def test_attr_transformation_fail(self):
        attrs = _Attrs('test', {'number': 'abc'}, {'count': ('number', int)})
        self.assertRaises(ValueError, lambda: attrs.count)

class ChildrenTest(TestCase):
    def setUp(self):
        self.dom = etree.fromstring(
            '<main><some name="0"/><any name="1"/><any name="2"/></main>'
        )

    def wrap(self, element):
        return '{0}.{1}'.format(element.tag, element.attrib['name'])

    def test_get_declared_section(self):
        children = _Children(
            'test', self.dom, {}, {'some_section': ('some', self.wrap)}
        )
        self.assertEqual('some.0', children.some_section)

    def test_get_declared_children(self):
        children = _Children('test', self.dom, {'anys': ('any', self.wrap)}, {})
        self.assertEqual(['any.1', 'any.2'], children.anys)

    def test_raises_on_undeclared_children(self):
        children = _Children('test', self.dom, {}, {})
        self.assertRaises(AttributeError, lambda: children.some_section)


class TestBase(TestCase):
    def setUp(self):
        self.create_covered_status = get_xml_manipulation_creator_from_file(
            rc('crm_mon.minimal.xml')
        )
        self.covered_status = self.create_covered_status()

class ClusterStatusTest(TestBase):
    def test_minimal_crm_mon_is_valid(self):
        ClusterState(str(self.covered_status))

    def test_refuse_invalid_xml(self):
        assert_raise_library_error(
            lambda: ClusterState('invalid xml'),
            (severities.ERROR, report_codes.BAD_CLUSTER_STATE_FORMAT, {})
        )

    def test_refuse_invalid_document(self):
        self.covered_status.append_to_first_tag_name(
            'nodes',
            '<node without="required attributes" />'
        )

        assert_raise_library_error(
            lambda: ClusterState(str(self.covered_status)),
            (severities.ERROR, report_codes.BAD_CLUSTER_STATE_FORMAT, {})
        )


class WorkWithClusterStatusNodesTest(TestBase):
    def fixture_node_string(self, **kwargs):
        attrs = dict(name='name', id='id', type='member')
        attrs.update(kwargs)
        return '''<node
            name="{name}"
            id="{id}"
            online="true"
            standby="true"
            standby_onfail="false"
            maintenance="false"
            pending="false"
            unclean="false"
            shutdown="false"
            expected_up="false"
            is_dc="false"
            resources_running="0"
            type="{type}"
        />'''.format(**attrs)

    def test_can_get_node_names(self):
        self.covered_status.append_to_first_tag_name(
            'nodes',
            self.fixture_node_string(name='node1', id='1'),
            self.fixture_node_string(name='node2', id='2'),
        )
        xml = str(self.covered_status)
        self.assertEqual(
            ['node1', 'node2'],
            [node.attrs.name for node in ClusterState(xml).node_section.nodes]
        )

    def test_can_filter_out_remote_nodes(self):
        self.covered_status.append_to_first_tag_name(
            'nodes',
            self.fixture_node_string(name='node1', id='1'),
            self.fixture_node_string(name='node2', type='remote', id='2'),
        )
        xml = str(self.covered_status)
        self.assertEqual(
            ['node1'],
            [
                node.attrs.name
                for node in ClusterState(xml).node_section.nodes
                if node.attrs.type != 'remote'
            ]
        )


class WorkWithClusterStatusSummaryTest(TestBase):
    def test_nodes_count(self):
        xml = str(self.covered_status)
        self.assertEqual(0, ClusterState(xml).summary.nodes.attrs.count)

    def test_resources_count(self):
        xml = str(self.covered_status)
        self.assertEqual(0, ClusterState(xml).summary.resources.attrs.count)


class GetPrimitiveRolesWithNodes(TestCase):
    def test_success(self):
        primitives_xml = [
            """
                <resource id="A" role="Started">
                    <node name="node1" id="1"/>
                </resource>
            """,
            """
                <resource id="A" role="Master">
                    <node name="node2" id="2"/>
                </resource>
            """,
            """
                <resource id="A" role="Slave">
                    <node name="node4" id="4"/>
                </resource>
            """,
            """
                <resource id="A" role="Slave">
                    <node name="node3" id="3"/>
                </resource>
            """,
            """
                <resource id="A" role="Stopped">
                </resource>
            """,
            """
                <resource id="A" role="Started">
                    <node name="node5" id="5"/>
                </resource>
            """,
        ]
        primitives = [
            etree.fromstring(xml) for xml in primitives_xml
        ]

        self.assertEqual(
            state._get_primitive_roles_with_nodes(primitives),
            {
                "Started": ["node1", "node5"],
                "Master": ["node2"],
                "Slave": ["node3", "node4"]
            }
        )

    def test_empty(self):
        self.assertEqual(
            state._get_primitive_roles_with_nodes([]),
            {
            }
        )


class GetPrimitivesForStateCheck(TestCase):
    # pylint: disable=too-many-public-methods
    status_xml = etree.fromstring("""
        <resources>
            <resource id="R01" failed="false" />
            <resource id="R02" failed="true" />

            <group id="G1">
                <resource id="R03" failed="false" />
                <resource id="R04" failed="false" />
            </group>
            <group id="G2">
                <resource id="R05" failed="true" />
                <resource id="R06" failed="true" />
            </group>

            <clone id="R07-clone">
                <resource id="R07" failed="false" />
                <resource id="R07" failed="false" />
            </clone>
            <clone id="R08-clone">
                <resource id="R08" failed="true" />
                <resource id="R08" failed="false" />
            </clone>
            <clone id="R09-clone">
                <resource id="R09" failed="true" />
                <resource id="R09" failed="true" />
            </clone>

            <clone id="R10-clone">
                <resource id="R10:0" failed="false" />
                <resource id="R10:1" failed="false" />
            </clone>
            <clone id="R11-clone">
                <resource id="R11:0" failed="false" />
                <resource id="R11:1" failed="true" />
            </clone>
            <clone id="R12-clone">
                <resource id="R12:0" failed="true" />
                <resource id="R12:1" failed="true" />
            </clone>

            <clone id="G3-clone">
                <group id="G3:0">
                    <resource id="R13" failed="false" />
                    <resource id="R14" failed="false" />
                </group>
                <group id="G3:1">
                    <resource id="R13" failed="false" />
                    <resource id="R14" failed="false" />
                </group>
            </clone>
            <clone id="G4-clone">
                <group id="G4:0">
                    <resource id="R15" failed="true" />
                    <resource id="R16" failed="true" />
                </group>
                <group id="G4:1">
                    <resource id="R15" failed="false" />
                    <resource id="R16" failed="false" />
                </group>
            </clone>
            <clone id="G5-clone">
                <group id="G5:0">
                    <resource id="R17" failed="true" />
                    <resource id="R18" failed="true" />
                </group>
                <group id="G5:1">
                    <resource id="R17" failed="true" />
                    <resource id="R18" failed="true" />
                </group>
            </clone>

            <clone id="G6-clone" managed="true">
                <group id="G6:0">
                    <resource id="R19:0" failed="false" />
                    <resource id="R20:0" failed="false" />
                </group>
                <group id="G6:1">
                    <resource id="R19:1" failed="false" />
                    <resource id="R20:1" failed="false" />
                </group>
            </clone>
            <clone id="G7-clone" managed="true">
                <group id="G7:0">
                    <resource id="R21:0" failed="true" />
                    <resource id="R22:0" failed="true" />
                </group>
                <group id="G7:1">
                    <resource id="R21:1" failed="false" />
                    <resource id="R22:1" failed="false" />
                </group>
            </clone>
            <clone id="G8-clone" managed="true">
                <group id="G8:0">
                    <resource id="R23:0" failed="true" />
                    <resource id="R24:0" failed="true" />
                </group>
                <group id="G8:1">
                    <resource id="R23:1" failed="true" />
                    <resource id="R24:1" failed="true" />
                </group>
            </clone>
            <bundle id="B1" managed="true">
                <replica id="0">
                    <resource id="B1-R1" failed="false" />
                    <resource id="B1-R2" failed="false" />
                </replica>
            </bundle>
            <bundle id="B2" managed="true">
                <replica id="0">
                    <resource id="B2-R1" failed="true" />
                    <resource id="B2-R2" failed="false" />
                </replica>
                <replica id="1">
                    <resource id="B2-R1" failed="false" />
                    <resource id="B2-R2" failed="false" />
                </replica>
            </bundle>
        </resources>
    """)

    def setUp(self):
        self.status = etree.parse(rc("crm_mon.minimal.xml")).getroot()
        self.status.append(self.status_xml)
        for resource in self.status.xpath(".//resource"):
            resource.attrib.update({
                "resource_agent": "ocf::pacemaker:Stateful",
                "role": "Started",
                "active": "true",
                "orphaned": "false",
                "blocked": "false",
                "managed": "true",
                "failure_ignored": "false",
                "nodes_running_on": "1",
            })

    def assert_primitives(self, resource_id, primitive_ids, expected_running):
        self.assertEqual(
            [
                elem.attrib["id"]
                for elem in state._get_primitives_for_state_check(
                    self.status, resource_id, expected_running
                )
            ],
            primitive_ids
        )

    def test_missing(self):
        self.assert_primitives("Rxx", [], True)
        self.assert_primitives("Rxx", [], False)

    def test_primitive(self):
        self.assert_primitives("R01", ["R01"], True)
        self.assert_primitives("R01", ["R01"], False)

    def test_primitive_failed(self):
        self.assert_primitives("R02", [], True)
        self.assert_primitives("R02", [], False)

    def test_group(self):
        self.assert_primitives("G1", ["R04"], True)
        self.assert_primitives("G1", ["R03"], False)

    def test_group_failed_primitive(self):
        self.assert_primitives("G2", [], True)
        self.assert_primitives("G2", [], False)

    def test_primitive_in_group(self):
        self.assert_primitives("R03", ["R03"], True)
        self.assert_primitives("R03", ["R03"], False)

    def test_primitive_in_group_failed(self):
        self.assert_primitives("R05", [], True)
        self.assert_primitives("R05", [], False)

    def test_clone(self):
        self.assert_primitives("R07-clone", ["R07", "R07"], True)
        self.assert_primitives("R07-clone", ["R07", "R07"], False)
        self.assert_primitives("R10-clone", ["R10:0", "R10:1"], True)
        self.assert_primitives("R10-clone", ["R10:0", "R10:1"], False)

    def test_clone_partially_failed(self):
        self.assert_primitives("R08-clone", ["R08"], True)
        self.assert_primitives("R08-clone", ["R08"], False)
        self.assert_primitives("R11-clone", ["R11:0"], True)
        self.assert_primitives("R11-clone", ["R11:0"], False)

    def test_clone_failed(self):
        self.assert_primitives("R09-clone", [], True)
        self.assert_primitives("R09-clone", [], False)
        self.assert_primitives("R12-clone", [], True)
        self.assert_primitives("R12-clone", [], False)

    def test_primitive_in_clone(self):
        self.assert_primitives("R07", ["R07", "R07"], True)
        self.assert_primitives("R07", ["R07", "R07"], False)
        self.assert_primitives("R10", ["R10:0", "R10:1"], True)
        self.assert_primitives("R10", ["R10:0", "R10:1"], False)

    def test_primitive_in_clone_partially_failed(self):
        self.assert_primitives("R08", ["R08"], True)
        self.assert_primitives("R08", ["R08"], False)
        self.assert_primitives("R11", ["R11:0"], True)
        self.assert_primitives("R11", ["R11:0"], False)

    def test_primitive_in_clone_failed(self):
        self.assert_primitives("R09", [], True)
        self.assert_primitives("R09", [], False)
        self.assert_primitives("R12", [], True)
        self.assert_primitives("R12", [], False)

    def test_clone_containing_group(self):
        self.assert_primitives("G3-clone", ["R14", "R14"], True)
        self.assert_primitives("G3-clone", ["R13", "R13"], False)
        self.assert_primitives("G6-clone", ["R20:0", "R20:1"], True)
        self.assert_primitives("G6-clone", ["R19:0", "R19:1"], False)

    def test_clone_containing_group_partially_failed(self):
        self.assert_primitives("G4-clone", ["R16"], True)
        self.assert_primitives("G4-clone", ["R15"], False)
        self.assert_primitives("G7-clone", ["R22:1"], True)
        self.assert_primitives("G7-clone", ["R21:1"], False)

    def test_clone_containing_group_failed(self):
        self.assert_primitives("G5-clone", [], True)
        self.assert_primitives("G5-clone", [], False)
        self.assert_primitives("G8-clone", [], True)
        self.assert_primitives("G8-clone", [], False)

    def test_group_in_clone_containing_group(self):
        self.assert_primitives("G3", ["R14", "R14"], True)
        self.assert_primitives("G3", ["R13", "R13"], False)
        self.assert_primitives("G6", ["R20:0", "R20:1"], True)
        self.assert_primitives("G6", ["R19:0", "R19:1"], False)

    def test_group_in_clone_containing_group_partially_failed(self):
        self.assert_primitives("G4", ["R16"], True)
        self.assert_primitives("G4", ["R15"], False)
        self.assert_primitives("G7", ["R22:1"], True)
        self.assert_primitives("G7", ["R21:1"], False)

    def test_group_in_clone_containing_group_failed(self):
        self.assert_primitives("G5", [], True)
        self.assert_primitives("G5", [], False)
        self.assert_primitives("G8", [], True)
        self.assert_primitives("G8", [], False)

    def test_primitive_in_clone_containing_group(self):
        self.assert_primitives("R14", ["R14", "R14"], True)
        self.assert_primitives("R14", ["R14", "R14"], False)
        self.assert_primitives("R20", ["R20:0", "R20:1"], True)
        self.assert_primitives("R20", ["R20:0", "R20:1"], False)

    def test_primitive_in_clone_containing_group_partially_failed(self):
        self.assert_primitives("R16", ["R16"], True)
        self.assert_primitives("R16", ["R16"], False)
        self.assert_primitives("R22", ["R22:1"], True)
        self.assert_primitives("R22", ["R22:1"], False)

    def test_primitive_in_clone_containing_group_failed(self):
        self.assert_primitives("R18", [], True)
        self.assert_primitives("R18", [], False)
        self.assert_primitives("R24", [], True)
        self.assert_primitives("R24", [], False)

    def test_bundle(self):
        self.assert_primitives("B1", ["B1-R1", "B1-R2"], True)
        self.assert_primitives("B1", ["B1-R1", "B1-R2"], False)
        self.assert_primitives("B2", ["B2-R2", "B2-R1", "B2-R2"], True)
        self.assert_primitives("B2", ["B2-R2", "B2-R1", "B2-R2"], False)

    def test_primitive_in_bundle(self):
        self.assert_primitives("B1-R1", ["B1-R1"], True)
        self.assert_primitives("B1-R1", ["B1-R1"], False)
        self.assert_primitives("B2-R1", ["B2-R1"], True)
        self.assert_primitives("B2-R1", ["B2-R1"], False)
        self.assert_primitives("B2-R2", ["B2-R2", "B2-R2"], True)
        self.assert_primitives("B2-R2", ["B2-R2", "B2-R2"], False)


class CommonResourceState(TestCase):
    resource_id = "R"
    def setUp(self):
        self.cluster_state = "state"

        patcher_primitives = mock.patch(
            "pcs.lib.pacemaker.state._get_primitives_for_state_check"
        )
        self.addCleanup(patcher_primitives.stop)
        self.get_primitives_for_state_check = patcher_primitives.start()

        patcher_roles = mock.patch(
            "pcs.lib.pacemaker.state._get_primitive_roles_with_nodes"
        )
        self.addCleanup(patcher_roles.stop)
        self.get_primitive_roles_with_nodes = patcher_roles.start()

    def fixture_running_state_info(self):
        return {
            "Started": ["node1"],
            "Master": ["node2"],
            "Slave": ["node3", "node4"],
        }

    def fixture_running_report(self, severity):
        return (severity, report_codes.RESOURCE_RUNNING_ON_NODES, {
            "resource_id": self.resource_id,
            "roles_with_nodes": self.fixture_running_state_info(),
        })

    def fixture_not_running_report(self, severity):
        return (severity, report_codes.RESOURCE_DOES_NOT_RUN, {
            "resource_id": self.resource_id
        })


class EnsureResourceState(CommonResourceState):
    def assert_running_info_transform(self, run_info, report, expected_running):
        self.get_primitives_for_state_check.return_value = ["elem1", "elem2"]
        self.get_primitive_roles_with_nodes.return_value = run_info
        assert_report_item_equal(
            state.ensure_resource_state(
                expected_running,
                self.cluster_state,
                self.resource_id
            ),
            report
        )
        self.get_primitives_for_state_check.assert_called_once_with(
            self.cluster_state,
            self.resource_id,
            expected_running
        )
        self.get_primitive_roles_with_nodes.assert_called_once_with(
            ["elem1", "elem2"]
        )

    def test_report_info_running(self):
        self.assert_running_info_transform(
            self.fixture_running_state_info(),
            self.fixture_running_report(severities.INFO),
            expected_running=True,
        )

    def test_report_error_running(self):
        self.assert_running_info_transform(
            self.fixture_running_state_info(),
            self.fixture_running_report(severities.ERROR),
            expected_running=False,
        )

    def test_report_error_not_running(self):
        self.assert_running_info_transform(
            [],
            self.fixture_not_running_report(severities.ERROR),
            expected_running=True,
        )

    def test_report_info_not_running(self):
        self.assert_running_info_transform(
            [],
            self.fixture_not_running_report(severities.INFO),
            expected_running=False,
        )


class InfoResourceState(CommonResourceState):
    def assert_running_info_transform(self, run_info, report):
        self.get_primitives_for_state_check.return_value = ["elem1", "elem2"]
        self.get_primitive_roles_with_nodes.return_value = run_info
        assert_report_item_equal(
            state.info_resource_state(self.cluster_state, self.resource_id),
            report
        )
        self.get_primitives_for_state_check.assert_called_once_with(
            self.cluster_state,
            self.resource_id,
            expected_running=True
        )
        self.get_primitive_roles_with_nodes.assert_called_once_with(
            ["elem1", "elem2"]
        )

    def test_report_info_running(self):
        self.assert_running_info_transform(
            self.fixture_running_state_info(),
            self.fixture_running_report(severities.INFO)
        )
    def test_report_info_not_running(self):
        self.assert_running_info_transform(
            [],
            self.fixture_not_running_report(severities.INFO)
        )


class IsResourceManaged(TestCase):
    status_xml = etree.fromstring("""
        <resources>
            <resource id="R01" managed="true" />
            <resource id="R02" managed="false" />

            <group id="G1">
                <resource id="R03" managed="true" />
                <resource id="R04" managed="true" />
            </group>
            <group id="G2">
                <resource id="R05" managed="false" />
                <resource id="R06" managed="true" />
            </group>
            <group id="G3">
                <resource id="R07" managed="true" />
                <resource id="R08" managed="false" />
            </group>
            <group id="G4">
                <resource id="R09" managed="false" />
                <resource id="R10" managed="false" />
            </group>

            <clone id="R11-clone" managed="true">
                <resource id="R11" managed="true" />
                <resource id="R11" managed="true" />
            </clone>
            <clone id="R12-clone" managed="true">
                <resource id="R12" managed="false" />
                <resource id="R12" managed="false" />
            </clone>
            <clone id="R13-clone" managed="false">
                <resource id="R13" managed="true" />
                <resource id="R13" managed="true" />
            </clone>
            <clone id="R14-clone" managed="false">
                <resource id="R14" managed="false" />
                <resource id="R14" managed="false" />
            </clone>

            <clone id="R15-clone" managed="true">
                <resource id="R15:0" managed="true" />
                <resource id="R15:1" managed="true" />
            </clone>
            <clone id="R16-clone" managed="true">
                <resource id="R16:0" managed="false" />
                <resource id="R16:1" managed="false" />
            </clone>
            <clone id="R17-clone" managed="false">
                <resource id="R17:0" managed="true" />
                <resource id="R17:1" managed="true" />
            </clone>
            <clone id="R18-clone" managed="false">
                <resource id="R18:0" managed="false" />
                <resource id="R18:1" managed="false" />
            </clone>

            <clone id="G5-clone" managed="true">
                <group id="G5:0">
                    <resource id="R19" managed="true" />
                    <resource id="R20" managed="true" />
                </group>
                <group id="G5:1">
                    <resource id="R19" managed="true" />
                    <resource id="R20" managed="true" />
                </group>
            </clone>
            <clone id="G6-clone" managed="false">
                <group id="G6:0">
                    <resource id="R21" managed="true" />
                    <resource id="R22" managed="true" />
                </group>
                <group id="G6:1">
                    <resource id="R21" managed="true" />
                    <resource id="R22" managed="true" />
                </group>
            </clone>
            <clone id="G7-clone" managed="true">
                <group id="G7:0">
                    <resource id="R23" managed="false" />
                    <resource id="R24" managed="true" />
                </group>
                <group id="G7:1">
                    <resource id="R23" managed="false" />
                    <resource id="R24" managed="true" />
                </group>
            </clone>
            <clone id="G8-clone" managed="true">
                <group id="G8:0">
                    <resource id="R25" managed="true" />
                    <resource id="R26" managed="false" />
                </group>
                <group id="G8:1">
                    <resource id="R25" managed="true" />
                    <resource id="R26" managed="false" />
                </group>
            </clone>
            <clone id="G9-clone" managed="false">
                <group id="G9:0">
                    <resource id="R27" managed="false" />
                    <resource id="R28" managed="false" />
                </group>
                <group id="G9:1">
                    <resource id="R27" managed="false" />
                    <resource id="R28" managed="false" />
                </group>
            </clone>

            <clone id="G10-clone" managed="true">
                <group id="G10:0">
                    <resource id="R29:0" managed="true" />
                    <resource id="R30:0" managed="true" />
                </group>
                <group id="G10:1">
                    <resource id="R29:1" managed="true" />
                    <resource id="R30:1" managed="true" />
                </group>
            </clone>
            <clone id="G11-clone" managed="false">
                <group id="G11:0">
                    <resource id="R31:0" managed="true" />
                    <resource id="R32:0" managed="true" />
                </group>
                <group id="G11:1">
                    <resource id="R31:1" managed="true" />
                    <resource id="R32:1" managed="true" />
                </group>
            </clone>
            <clone id="G12-clone" managed="true">
                <group id="G12:0">
                    <resource id="R33:0" managed="false" />
                    <resource id="R34:0" managed="true" />
                </group>
                <group id="G12:1">
                    <resource id="R33:1" managed="false" />
                    <resource id="R34:1" managed="true" />
                </group>
            </clone>
            <clone id="G13-clone" managed="true">
                <group id="G13:0">
                    <resource id="R35:0" managed="true" />
                    <resource id="R36:0" managed="false" />
                </group>
                <group id="G13:1">
                    <resource id="R35:1" managed="true" />
                    <resource id="R36:1" managed="false" />
                </group>
            </clone>
            <clone id="G14-clone" managed="false">
                <group id="G14:0">
                    <resource id="R37:0" managed="false" />
                    <resource id="R38:0" managed="false" />
                </group>
                <group id="G14:1">
                    <resource id="R37:1" managed="false" />
                    <resource id="R38:1" managed="false" />
                </group>
            </clone>

            <bundle id="B1" managed="true" />
            <bundle id="B2" managed="false" />

            <bundle id="B3" managed="true">
                <replica id="0">
                    <resource id="R39" managed="true" />
                    <resource id="R40" managed="true" />
                </replica>
                <replica id="1">
                    <resource id="R39" managed="true" />
                    <resource id="R40" managed="true" />
                </replica>
            </bundle>
            <bundle id="B4" managed="false">
                <replica id="0">
                    <resource id="R41" managed="true" />
                    <resource id="R42" managed="true" />
                </replica>
                <replica id="1">
                    <resource id="R41" managed="true" />
                    <resource id="R42" managed="true" />
                </replica>
            </bundle>
            <bundle id="B5" managed="true">
                <replica id="0">
                    <resource id="R43" managed="false" />
                    <resource id="R44" managed="true" />
                </replica>
                <replica id="1">
                    <resource id="R43" managed="false" />
                    <resource id="R44" managed="true" />
                </replica>
            </bundle>
            <bundle id="B6" managed="true">
                <replica id="0">
                    <resource id="R45" managed="true" />
                    <resource id="R46" managed="false" />
                </replica>
                <replica id="1">
                    <resource id="R45" managed="true" />
                    <resource id="R46" managed="false" />
                </replica>
            </bundle>
            <bundle id="B7" managed="false">
                <replica id="0">
                    <resource id="R47" managed="false" />
                    <resource id="R48" managed="false" />
                </replica>
                <replica id="1">
                    <resource id="R47" managed="false" />
                    <resource id="R48" managed="false" />
                </replica>
            </bundle>
        </resources>
    """)

    def setUp(self):
        self.status = etree.parse(rc("crm_mon.minimal.xml")).getroot()
        self.status.append(self.status_xml)
        for resource in self.status.xpath(".//resource"):
            resource.attrib.update({
                "resource_agent": "ocf::pacemaker:Stateful",
                "role": "Started",
                "active": "true",
                "orphaned": "false",
                "blocked": "false",
                "failed": "false",
                "failure_ignored": "false",
                "nodes_running_on": "1",
            })

    def assert_managed(self, resource, managed):
        self.assertEqual(
            managed,
            state.is_resource_managed(self.status, resource)
        )

    def test_missing(self):
        self.assertRaises(
            state.ResourceNotFound,
            self.assert_managed, "Rxx", True
        )

    def test_primitive(self):
        self.assert_managed("R01", True)
        self.assert_managed("R02", False)

    def test_group(self):
        self.assert_managed("G1", True)
        self.assert_managed("G2", False)
        self.assert_managed("G3", False)
        self.assert_managed("G4", False)

    def test_primitive_in_group(self):
        self.assert_managed("R03", True)
        self.assert_managed("R04", True)
        self.assert_managed("R05", False)
        self.assert_managed("R06", True)
        self.assert_managed("R07", True)
        self.assert_managed("R08", False)
        self.assert_managed("R09", False)
        self.assert_managed("R10", False)

    def test_clone(self):
        self.assert_managed("R11-clone", True)
        self.assert_managed("R12-clone", False)
        self.assert_managed("R13-clone", False)
        self.assert_managed("R14-clone", False)

        self.assert_managed("R15-clone", True)
        self.assert_managed("R16-clone", False)
        self.assert_managed("R17-clone", False)
        self.assert_managed("R18-clone", False)

    def test_primitive_in_clone(self):
        self.assert_managed("R11", True)
        self.assert_managed("R12", False)
        self.assert_managed("R13", False)
        self.assert_managed("R14", False)

    def test_primitive_in_unique_clone(self):
        self.assert_managed("R15", True)
        self.assert_managed("R16", False)
        self.assert_managed("R17", False)
        self.assert_managed("R18", False)

    def test_clone_containing_group(self):
        self.assert_managed("G5-clone", True)
        self.assert_managed("G6-clone", False)
        self.assert_managed("G7-clone", False)
        self.assert_managed("G8-clone", False)
        self.assert_managed("G9-clone", False)

        self.assert_managed("G10-clone", True)
        self.assert_managed("G11-clone", False)
        self.assert_managed("G12-clone", False)
        self.assert_managed("G13-clone", False)
        self.assert_managed("G14-clone", False)

    def test_group_in_clone(self):
        self.assert_managed("G5", True)
        self.assert_managed("G6", False)
        self.assert_managed("G7", False)
        self.assert_managed("G8", False)
        self.assert_managed("G9", False)

    def test_group_in_unique_clone(self):
        self.assert_managed("G10", True)
        self.assert_managed("G11", False)
        self.assert_managed("G12", False)
        self.assert_managed("G13", False)
        self.assert_managed("G14", False)

    def test_primitive_in_group_in_clone(self):
        self.assert_managed("R19", True)
        self.assert_managed("R20", True)
        self.assert_managed("R21", False)
        self.assert_managed("R22", False)
        self.assert_managed("R23", False)
        self.assert_managed("R24", True)
        self.assert_managed("R25", True)
        self.assert_managed("R26", False)
        self.assert_managed("R27", False)
        self.assert_managed("R28", False)

    def test_primitive_in_group_in_unique_clone(self):
        self.assert_managed("R29", True)
        self.assert_managed("R30", True)
        self.assert_managed("R31", False)
        self.assert_managed("R32", False)
        self.assert_managed("R33", False)
        self.assert_managed("R34", True)
        self.assert_managed("R35", True)
        self.assert_managed("R36", False)
        self.assert_managed("R37", False)
        self.assert_managed("R38", False)

    def test_bundle(self):
        self.assert_managed("B1", True)
        self.assert_managed("B2", False)
        self.assert_managed("B3", True)
        self.assert_managed("B4", False)
        self.assert_managed("B5", False)
        self.assert_managed("B6", False)
        self.assert_managed("B7", False)

    def test_primitive_in_bundle(self):
        self.assert_managed("R39", True)
        self.assert_managed("R40", True)
        self.assert_managed("R41", False)
        self.assert_managed("R42", False)
        self.assert_managed("R43", False)
        self.assert_managed("R44", True)
        self.assert_managed("R45", True)
        self.assert_managed("R46", False)
        self.assert_managed("R47", False)
        self.assert_managed("R48", False)



import os.path
from unittest import mock, TestCase

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal

from pcs import settings
from pcs.common import file_type_codes, report_codes
from pcs.common.file import RawFileError
from pcs.lib.booth import config_files
from pcs.lib.booth.config_facade import ConfigFacade


@mock.patch("os.path.isdir")
@mock.patch("os.listdir")
@mock.patch("os.path.isfile")
class GetAllConfigsFileNamesTest(TestCase):
    def test_booth_config_dir_is_no_dir(
        self, mock_is_file, mock_listdir, mock_isdir
    ):
        mock_isdir.return_value = False
        self.assertEqual([], config_files.get_all_configs_file_names())
        mock_isdir.assert_called_once_with(settings.booth_config_dir)
        self.assertEqual(0, mock_is_file.call_count)
        self.assertEqual(0, mock_listdir.call_count)

    def test_success(self, mock_is_file, mock_listdir, mock_isdir):
        def mock_is_file_fn(file_name):
            if file_name in [
                os.path.join(settings.booth_config_dir, name)
                for name in ("dir.cong", "dir")
            ]:
                return False
            if file_name in [
                os.path.join(settings.booth_config_dir, name)
                for name in (
                    "name1", "name2.conf", "name.conf.conf", ".conf",
                    "name3.conf"
                )
            ]:
                return True

            raise AssertionError("unexpected input")

        mock_isdir.return_value = True
        mock_is_file.side_effect = mock_is_file_fn
        mock_listdir.return_value = [
            "name1", "name2.conf", "name.conf.conf", ".conf", "name3.conf",
            "dir.cong", "dir"
        ]
        self.assertEqual(
            ["name2.conf", "name.conf.conf", "name3.conf"],
            config_files.get_all_configs_file_names()
        )
        mock_listdir.assert_called_once_with(settings.booth_config_dir)


class GetAuthfileNameAndData(TestCase):
    def test_no_auth_file(self):
        conf = ConfigFacade.create([], [])
        self.assertEqual(
            (None, None, []),
            config_files.get_authfile_name_and_data(conf)
        )

    def assert_path_check(self, path):
        conf = ConfigFacade.create([], [])
        conf.set_authfile(path)
        result = config_files.get_authfile_name_and_data(conf)
        self.assertEqual(result[0], None)
        self.assertEqual(result[1], None)
        assert_report_item_list_equal(
            result[2],
            [
                fixture.warn(
                    report_codes.BOOTH_UNSUPPORTED_FILE_LOCATION,
                    file_type_code=file_type_codes.BOOTH_KEY,
                    file_path=path,
                    expected_dir=settings.booth_config_dir,
                ),
            ]
        )

    def test_invalid_path(self):
        self.assert_path_check("/not/etc/booth/booth.key")

    def test_not_abs_path(self):
        self.assert_path_check("/etc/booth/../booth.key")

    @mock.patch("pcs.common.file.RawFile.read")
    def test_success(self, mock_read):
        key_data = "some key data".encode("utf-8")
        key_name = "my_booth.key"
        mock_read.return_value = key_data
        path = os.path.join(settings.booth_config_dir, key_name)
        conf = ConfigFacade.create([], [])
        conf.set_authfile(path)
        self.assertEqual(
            (key_name, key_data, []),
            config_files.get_authfile_name_and_data(conf)
        )

    @mock.patch("pcs.common.file.RawFile.read")
    def test_read_failure(self, mock_read):
        key_name = "my_booth.key"
        mock_read.side_effect = RawFileError(
            "mock file type",
            RawFileError.ACTION_READ,
            "some error"
        )
        path = os.path.join(settings.booth_config_dir, key_name)
        conf = ConfigFacade.create([], [])
        conf.set_authfile(path)
        with self.assertRaises(RawFileError):
            config_files.get_authfile_name_and_data(conf)



from unittest import TestCase

from pcs.common.pacemaker.resource import relations


class RelationEntityDto(TestCase):
    def setUp(self):
        self.an_id = "an_id"
        self.members = ["m2", "m1", "m3", "m0"]
        self.metadata = dict(
            key1="val1",
            key0="vel0",
            keyx="valx",
        )
        self.str_type = "a_type"
        self.enum_type = relations.ResourceRelationType.ORDER_SET

    def dto_fixture(self, a_type):
        return relations.RelationEntityDto(
            self.an_id, a_type, self.members, self.metadata
        )

    def dict_fixture(self, a_type):
        return dict(
            id=self.an_id,
            type=a_type,
            members=self.members,
            metadata=self.metadata,
        )

    def test_to_dict_type_str(self):
        dto = self.dto_fixture(self.str_type)
        self.assertEqual(self.dict_fixture(self.str_type), dto.to_dict())

    def test_to_dict_type_enum(self):
        dto = self.dto_fixture(self.enum_type)
        self.assertEqual(self.dict_fixture(self.enum_type.value), dto.to_dict())

    def test_from_dict_type_str(self):
        dto = relations.RelationEntityDto.from_dict(
            self.dict_fixture(self.str_type)
        )
        self.assertEqual(dto.id, self.an_id)
        self.assertEqual(dto.type, self.str_type)
        self.assertEqual(dto.members, self.members)
        self.assertEqual(dto.metadata, self.metadata)

    def test_from_dict_type_enum(self):
        dto = relations.RelationEntityDto.from_dict(
            self.dict_fixture(self.enum_type.value)
        )
        self.assertEqual(dto.id, self.an_id)
        self.assertEqual(dto.type, self.enum_type)
        self.assertEqual(dto.members, self.members)
        self.assertEqual(dto.metadata, self.metadata)


class ResourceRelationDto(TestCase):
    @staticmethod
    def ent_dto_fixture(an_id):
        return relations.RelationEntityDto(
            an_id, "a_type", ["m1", "m0", "m2"], dict(k1="v1", kx="vx")
        )

    def test_to_dict_no_members(self):
        ent_dto = self.ent_dto_fixture("an_id")
        dto = relations.ResourceRelationDto(ent_dto, [], True)
        self.assertEqual(
            dict(
                relation_entity=ent_dto.to_dict(),
                members=[],
                is_leaf=True,
            ),
            dto.to_dict()
        )

    def test_to_dict_with_members(self):
        ent_dto = self.ent_dto_fixture("an_id")
        m1_ent = self.ent_dto_fixture("m1_ent")
        m2_ent = self.ent_dto_fixture("m2_ent")
        m3_ent = self.ent_dto_fixture("m3_ent")
        members = [
            relations.ResourceRelationDto(
                m1_ent, [], False
            ),
            relations.ResourceRelationDto(
                m2_ent,
                [relations.ResourceRelationDto(m3_ent, [], True)],
                False
            ),
        ]
        dto = relations.ResourceRelationDto(ent_dto, members, True)
        self.assertEqual(
            dict(
                relation_entity=ent_dto.to_dict(),
                members=[
                    dict(
                        relation_entity=m1_ent.to_dict(),
                        members=[],
                        is_leaf=False,
                    ),
                    dict(
                        relation_entity=m2_ent.to_dict(),
                        members=[
                            dict(
                                relation_entity=m3_ent.to_dict(),
                                members=[],
                                is_leaf=True,
                            )
                        ],
                        is_leaf=False,
                    ),
                ],
                is_leaf=True,
            ),
            dto.to_dict()
        )

    def test_from_dict(self):
        ent_dto = self.ent_dto_fixture("an_id")
        m1_ent = self.ent_dto_fixture("m1_ent")
        m2_ent = self.ent_dto_fixture("m2_ent")
        m3_ent = self.ent_dto_fixture("m3_ent")
        dto = relations.ResourceRelationDto.from_dict(
            dict(
                relation_entity=ent_dto.to_dict(),
                members=[
                    dict(
                        relation_entity=m1_ent.to_dict(),
                        members=[],
                        is_leaf=False,
                    ),
                    dict(
                        relation_entity=m2_ent.to_dict(),
                        members=[
                            dict(
                                relation_entity=m3_ent.to_dict(),
                                members=[],
                                is_leaf=True,
                            )
                        ],
                        is_leaf=False,
                    ),
                ],
                is_leaf=True,
            )
        )
        self.assertEqual(ent_dto.to_dict(), dto.relation_entity.to_dict())
        self.assertEqual(True, dto.is_leaf)
        self.assertEqual(2, len(dto.members))

        self.assertEqual(
            m1_ent.to_dict(), dto.members[0].relation_entity.to_dict()
        )
        self.assertEqual(False, dto.members[0].is_leaf)
        self.assertEqual(0, len(dto.members[0].members))

        self.assertEqual(
            m2_ent.to_dict(), dto.members[1].relation_entity.to_dict()
        )
        self.assertEqual(False, dto.members[1].is_leaf)
        self.assertEqual(1, len(dto.members[1].members))

        self.assertEqual(
            m3_ent.to_dict(),
            dto.members[1].members[0].relation_entity.to_dict(),
        )
        self.assertEqual(True, dto.members[1].members[0].is_leaf)
        self.assertEqual(0, len(dto.members[1].members[0].members))



from collections import namedtuple
from functools import partial
from unittest import TestCase

from pcs.cli.common.reports import (
    CODE_BUILDER_MAP,
    build_message_from_report,
)
from pcs.common import report_codes

ReportItemMock = namedtuple("ReportItemMock", "code info")


class ReportsTranslated(TestCase):
    force_codes = {
        report_codes.FORCE,
        report_codes.FORCE_ALERT_RECIPIENT_VALUE_NOT_UNIQUE,
        report_codes.FORCE_ALREADY_IN_CLUSTER,
        report_codes.FORCE_BOOTH_DESTROY,
        report_codes.FORCE_BOOTH_REMOVE_FROM_CIB,
        report_codes.FORCE_REMOVE_MULTIPLE_NODES,
        report_codes.FORCE_CONSTRAINT_DUPLICATE,
        report_codes.FORCE_CONSTRAINT_MULTIINSTANCE_RESOURCE,
        report_codes.FORCE_FILE_OVERWRITE,
        report_codes.FORCE_LOAD_NODES_FROM_CIB,
        report_codes.FORCE_LOAD_THRESHOLD,
        report_codes.FORCE_METADATA_ISSUE,
        report_codes.FORCE_NODE_ADDRESSES_UNRESOLVABLE,
        report_codes.FORCE_NODE_DOES_NOT_EXIST,
        report_codes.FORCE_OPTIONS,
        report_codes.FORCE_QDEVICE_MODEL,
        report_codes.FORCE_QDEVICE_USED,
        report_codes.FORCE_QUORUM_LOSS,
        report_codes.FORCE_STONITH_RESOURCE_DOES_NOT_EXIST,
        report_codes.FORCE_NOT_SUITABLE_COMMAND,
        report_codes.FORCE_CLEAR_CLUSTER_NODE,
        report_codes.FORCE_RESOURCE_IN_BUNDLE_NOT_ACCESSIBLE,
        report_codes.SKIP_OFFLINE_NODES,
        report_codes.SKIP_FILE_DISTRIBUTION_ERRORS,
        report_codes.SKIP_ACTION_ON_NODES_ERRORS,
        report_codes.SKIP_UNREADABLE_CONFIG,
    }

    def test_all_reports_translated(self):
        all_codes = frozenset((
            getattr(report_codes, code_const)
            for code_const in dir(report_codes)
            if (
                not code_const.startswith("_") # skip python builtins
                and
                getattr(report_codes, code_const) not in self.force_codes
            )
        ))
        untranslated = all_codes - frozenset(CODE_BUILDER_MAP.keys())
        self.assertEqual(
            untranslated,
            frozenset(),
            f"{len(untranslated)} report codes have no translation in CLI. "
            "Add translations for the report codes. Add force codes to "
            f"{self.__class__.__module__}.{self.__class__.__name__}."
            "force_codes so they won't get reported here."
        )


class BuildMessageFromReportTest(TestCase):
    def test_returns_default_message_when_code_not_in_map(self):
        info = {"first": "FIRST"}
        self.assertEqual(
            "Unknown report: SOME info: {0}force text".format(str(info)),
            build_message_from_report(
                {},
                ReportItemMock("SOME", info),
                "force text"
            )
        )

    def test_complete_force_text(self):
        self.assertEqual(
            "Message force text is inside",
            build_message_from_report(
                {
                    "SOME": lambda info, force_text:
                        "Message "+force_text+" is inside"
                    ,
                },
                ReportItemMock("SOME", {}),
                "force text"
            )
        )

    def test_deal_with_callable(self):
        self.assertEqual(
            "Info: MESSAGE",
            build_message_from_report(
                {
                    "SOME": lambda info: "Info: {message}".format(**info),
                },
                ReportItemMock("SOME", {"message": "MESSAGE"}),
            )
        )

    def test_append_force_when_needed_and_not_specified(self):
        self.assertEqual(
            "message force at the end",
            build_message_from_report(
                {"SOME": "message"},
                ReportItemMock("SOME", {}),
                " force at the end",
            )
        )

    def test_returns_default_message_when_conflict_key_appear(self):
        info = {"message": "MESSAGE"}
        self.assertEqual(
            "Unknown report: SOME info: {0}".format(str(info)),
            build_message_from_report(
                {
                    "SOME": lambda info: "Info: {message} {extra}".format(
                        message="ANY", **info
                    ),
                },
                ReportItemMock("SOME", info),
            )
        )

    def test_returns_default_message_when_key_disappear(self):
        self.assertEqual(
            "Unknown report: SOME info: {}"
            ,
            build_message_from_report(
                {
                    "SOME": lambda info: "Info: {message}".format(**info),
                },
                ReportItemMock("SOME", {}),
            )
        )

    def test_callable_is_partial_object(self):
        code_builder_map = {
            "SOME": partial(
                lambda title, info: "{title}: {message}".format(
                    title=title, **info
                ),
                "Info"
            )
        }
        self.assertEqual(
            "Info: MESSAGE",
            build_message_from_report(
                code_builder_map,
                ReportItemMock("SOME", {"message": "MESSAGE"})
            )
        )

    def test_callable_is_partial_object_with_force(self):
        code_builder_map = {
            "SOME": partial(
                lambda title, info, force_text:
                    "{title}: {message} {force_text}".format(
                        title=title, force_text=force_text, **info
                    ),
                "Info"
            )
        }
        self.assertEqual(
            "Info: MESSAGE force text",
            build_message_from_report(
                code_builder_map,
                ReportItemMock("SOME", {"message": "MESSAGE"}),
                "force text"
            )
        )
